"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import os
from typing import OrderedDict
import copy
from fuse.dl.lightning.pl_funcs import convert_predictions_to_dataframe


import torch
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data.dataloader import DataLoader
import pytorch_lightning as pl

from fuse.eval.evaluator import EvaluatorDefault
from fuse.eval.metrics.classification.metrics_thresholding_common import MetricApplyThresholds
from fuse.eval.metrics.classification.metrics_classification_common import MetricAccuracy, MetricAUCROC, MetricROCCurve

from fuse.data.utils.samplers import BatchSamplerDefault
from fuse.data.utils.collates import CollateDefault
from fuse.data.utils.split import dataset_balanced_division_to_folds

from fuse.dl.losses.loss_default import LossDefault
from fuse.dl.models.backbones.backbone_resnet_3d import BackboneResnet3D
from fuse.dl.models import ModelMultiHead
from fuse.dl.models.heads.heads_3D import Head3DClassifier
from fuse.dl.lightning.pl_module import LightningModuleDefault

from fuse.utils.utils_debug import FuseDebug
from fuse.utils.utils_logger import fuse_logger_start
from fuse.utils.file_io.file_io import create_dir, load_pickle, save_dataframe
import fuse.utils.gpu as GPU

from fuseimg.datasets.stoic21 import STOIC21

###########################################################################################################
# Fuse
###########################################################################################################
##########################################
# Debug modes
##########################################
mode = "default"  # Options: 'default', 'debug'. See details in FuseDebug
debug = FuseDebug(mode)

##########################################qQ
# Output Paths
##########################################
assert (
    "STOIC21_DATA_PATH" in os.environ
), "Expecting environment variable STOIC21_DATA_PATH to be set. Follow the instruction in example README file to download and set the path to the data"
ROOT = "_examples/stoic21"  # TODO: fill path here
model_dir = os.path.join(ROOT, "model_dir")
PATHS = {
    "model_dir": model_dir,
    "cache_dir": os.path.join(ROOT, "cache_dir"),
    "data_split_filename": os.path.join(ROOT, "stoic21_split.pkl"),
    "data_dir": os.environ["STOIC21_DATA_PATH"],
    "inference_dir": os.path.join(model_dir, "infer_dir"),
    "eval_dir": os.path.join(model_dir, "eval_dir"),
}
NUM_GPUS = 1

##########################################
# Train Common Params
##########################################
TRAIN_COMMON_PARAMS = {}
# ============
# Model
# ============
TRAIN_COMMON_PARAMS["model"] = dict(imaging_dropout=0.5, fused_dropout=0.0, clinical_dropout=0.0)

# ============
# Data
# ============
TRAIN_COMMON_PARAMS["data.batch_size"] = 4
TRAIN_COMMON_PARAMS["data.train_num_workers"] = 16
TRAIN_COMMON_PARAMS["data.validation_num_workers"] = 16
TRAIN_COMMON_PARAMS["data.num_folds"] = 5
TRAIN_COMMON_PARAMS["data.train_folds"] = [0, 1, 2, 3]
TRAIN_COMMON_PARAMS["data.validation_folds"] = [4]

# ===============
# PL Trainer
# ===============
TRAIN_COMMON_PARAMS["trainer.num_epochs"] = 50
TRAIN_COMMON_PARAMS["trainer.num_devices"] = NUM_GPUS
TRAIN_COMMON_PARAMS["trainer.accelerator"] = "gpu"
# use "dp" strategy temp when working with multiple GPUS - workaround for pytorch lightning issue: https://github.com/Lightning-AI/lightning/issues/11807
TRAIN_COMMON_PARAMS["trainer.strategy"] = "dp" if TRAIN_COMMON_PARAMS["trainer.num_devices"] > 1 else None
TRAIN_COMMON_PARAMS["trainer.ckpt_path"] = None  # path to the checkpoint you wish continue the training from

# ===============
# Optimizer
# ===============
TRAIN_COMMON_PARAMS["opt.lr"] = 1e-3
TRAIN_COMMON_PARAMS["opt.weight_decay"] = 0.005


def create_model(imaging_dropout: float, clinical_dropout: float, fused_dropout: float) -> torch.nn.Module:
    """
    creates the model
    See Head3DClassifier for details about imaging_dropout, clinical_dropout, fused_dropout
    """
    model = ModelMultiHead(
        conv_inputs=(("data.input.img", 1),),
        backbone=BackboneResnet3D(in_channels=1),
        heads=[
            Head3DClassifier(
                head_name="classification",
                conv_inputs=[("model.backbone_features", 512)],
                dropout_rate=imaging_dropout,
                append_dropout_rate=clinical_dropout,
                fused_dropout_rate=fused_dropout,
                num_classes=2,
                append_features=[("data.input.clinical", 8)],
                append_layers_description=(256, 128),
            ),
        ],
    )
    return model


#################################
# Train Template
#################################
def run_train(paths: dict, train_common_params: dict):
    # ==============================================================================
    # Logger
    # ==============================================================================
    fuse_logger_start(output_path=paths["model_dir"], console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Train", {"attrs": ["bold", "underline"]})

    lgr.info(f'model_dir={paths["model_dir"]}', {"color": "magenta"})
    lgr.info(f'cache_dir={paths["cache_dir"]}', {"color": "magenta"})

    # ==============================================================================
    # Data
    # ==============================================================================
    # Train Data
    lgr.info("Train Data:", {"attrs": "bold"})

    # split to folds randomly - temp
    dataset_all = STOIC21.dataset(paths["data_dir"], paths["cache_dir"], reset_cache=False)
    folds = dataset_balanced_division_to_folds(
        dataset=dataset_all,
        output_split_filename=paths["data_split_filename"],
        keys_to_balance=["data.gt.probSevere"],
        nfolds=train_common_params["data.num_folds"],
    )

    train_sample_ids = []
    for fold in train_common_params["data.train_folds"]:
        train_sample_ids += folds[fold]
    validation_sample_ids = []
    for fold in train_common_params["data.validation_folds"]:
        validation_sample_ids += folds[fold]

    train_dataset = STOIC21.dataset(paths["data_dir"], paths["cache_dir"], sample_ids=train_sample_ids, train=True)
    validation_dataset = STOIC21.dataset(
        paths["data_dir"], paths["cache_dir"], sample_ids=validation_sample_ids, train=False
    )

    lgr.info("- Create sampler:")
    sampler = BatchSamplerDefault(
        dataset=train_dataset,
        balanced_class_name="data.gt.probSevere",
        num_balanced_classes=2,
        batch_size=train_common_params["data.batch_size"],
        balanced_class_weights=None,
    )
    lgr.info("- Create sampler: Done")

    # Create dataloader
    train_dataloader = DataLoader(
        dataset=train_dataset,
        batch_sampler=sampler,
        collate_fn=CollateDefault(),
        num_workers=train_common_params["data.train_num_workers"],
    )
    lgr.info("Train Data: Done", {"attrs": "bold"})

    # dataloader
    validation_dataloader = DataLoader(
        dataset=validation_dataset,
        batch_size=train_common_params["data.batch_size"],
        collate_fn=CollateDefault(),
        num_workers=train_common_params["data.validation_num_workers"],
    )
    lgr.info("Validation Data: Done", {"attrs": "bold"})

    # ==============================================================================
    # Model
    # ==============================================================================
    model = create_model(**train_common_params["model"])

    # ====================================================================================
    #  Loss
    # ====================================================================================
    losses = {
        "cls_loss": LossDefault(
            pred="model.logits.classification", target="data.gt.probSevere", callable=F.cross_entropy, weight=1.0
        ),
    }

    # ====================================================================================
    # Metrics
    # ====================================================================================
    train_metrics = OrderedDict(
        [
            ("auc", MetricAUCROC(pred="model.output.classification", target="data.gt.probSevere")),
        ]
    )

    validation_metrics = copy.deepcopy(train_metrics)  # use the same metrics in validation as well

    # either a dict with arguments to pass to ModelCheckpoint or list dicts for multiple ModelCheckpoint callbacks (to monitor and save checkpoints for more then one metric).
    best_epoch_source = dict(
        monitor="validation.metrics.auc",
        mode="max",
    )

    # ====================================================================================
    # Training components
    # ====================================================================================
    # create optimizer
    optimizer = optim.SGD(
        model.parameters(),
        lr=train_common_params["opt.lr"],
        weight_decay=train_common_params["opt.weight_decay"],
        momentum=0.9,
        nesterov=True,
    )

    # create learning scheduler
    lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer)
    lr_sch_config = dict(scheduler=lr_scheduler, monitor="validation.losses.total_loss")

    # optimizier and lr sch - see pl.LightningModule.configure_optimizers return value for all options
    optimizers_and_lr_schs = dict(optimizer=optimizer, lr_scheduler=lr_sch_config)

    # =====================================================================================
    #  Train
    # =====================================================================================
    lgr.info("Train:", {"attrs": "bold"})

    # create instance of PL module - FuseMedML generic version
    pl_module = LightningModuleDefault(
        model_dir=paths["model_dir"],
        model=model,
        losses=losses,
        train_metrics=train_metrics,
        validation_metrics=validation_metrics,
        best_epoch_source=best_epoch_source,
        optimizers_and_lr_schs=optimizers_and_lr_schs,
    )

    # create lightining trainer.
    pl_trainer = pl.Trainer(
        default_root_dir=paths["model_dir"],
        max_epochs=train_common_params["trainer.num_epochs"],
        accelerator=train_common_params["trainer.accelerator"],
        devices=train_common_params["trainer.num_devices"],
        strategy=train_common_params["trainer.strategy"],
        auto_select_gpus=True,
    )

    # train
    pl_trainer.fit(
        pl_module, train_dataloader, validation_dataloader, ckpt_path=train_common_params["trainer.ckpt_path"]
    )

    lgr.info("Train: Done", {"attrs": "bold"})


######################################
# Inference Common Params
######################################
INFER_COMMON_PARAMS = {}
INFER_COMMON_PARAMS["infer_filename"] = "infer_file.gz"
INFER_COMMON_PARAMS["checkpoint"] = "best_epoch.ckpt"
INFER_COMMON_PARAMS["data.infer_folds"] = [4]  # infer validation set
INFER_COMMON_PARAMS["data.batch_size"] = 4
INFER_COMMON_PARAMS["data.num_workers"] = 16
INFER_COMMON_PARAMS["model"] = TRAIN_COMMON_PARAMS["model"]
INFER_COMMON_PARAMS["trainer.num_devices"] = 1  # infer must use single device
INFER_COMMON_PARAMS["trainer.accelerator"] = "gpu"
INFER_COMMON_PARAMS["trainer.strategy"] = None

######################################
# Inference Template
######################################


def run_infer(paths: dict, infer_common_params: dict):
    create_dir(paths["inference_dir"])
    infer_file = os.path.join(paths["inference_dir"], infer_common_params["infer_filename"])
    checkpoint_file = os.path.join(paths["model_dir"], infer_common_params["checkpoint"])
    #### Logger
    fuse_logger_start(output_path=paths["inference_dir"], console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Inference", {"attrs": ["bold", "underline"]})
    lgr.info(f"infer_filename={infer_file}", {"color": "magenta"})

    ## Data
    folds = load_pickle(paths["data_split_filename"])  # assume exists and created in train func

    infer_sample_ids = []
    for fold in infer_common_params["data.infer_folds"]:
        infer_sample_ids += folds[fold]

    infer_dataset = STOIC21.dataset(paths["data_dir"], paths["cache_dir"], sample_ids=infer_sample_ids, train=False)
    infer_dataloader = DataLoader(dataset=infer_dataset, collate_fn=CollateDefault(), batch_size=2, num_workers=2)

    # load python lightning module
    model = create_model(**infer_common_params["model"])
    pl_module = LightningModuleDefault.load_from_checkpoint(
        checkpoint_file, model_dir=paths["model_dir"], model=model, map_location="cpu", strict=True
    )
    # set the prediction keys to extract (the ones used be the evaluation function).
    pl_module.set_predictions_keys(
        ["model.output.classification", "data.gt.probSevere"]
    )  # which keys to extract and dump into file

    # create a trainer instance
    pl_trainer = pl.Trainer(
        default_root_dir=paths["model_dir"],
        accelerator=infer_common_params["trainer.accelerator"],
        devices=infer_common_params["trainer.num_devices"],
        strategy=infer_common_params["trainer.strategy"],
        auto_select_gpus=True,
    )
    predictions = pl_trainer.predict(pl_module, infer_dataloader, return_predictions=True)

    # convert list of batch outputs into a dataframe
    infer_df = convert_predictions_to_dataframe(predictions)
    save_dataframe(infer_df, infer_file)


######################################
# Eval Common Params
######################################
EVAL_COMMON_PARAMS = {}
EVAL_COMMON_PARAMS["infer_filename"] = INFER_COMMON_PARAMS["infer_filename"]


######################################
# Eval Template
######################################
def run_eval(paths: dict, eval_common_params: dict):
    infer_file = os.path.join(paths["inference_dir"], eval_common_params["infer_filename"])

    fuse_logger_start(output_path=None, console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Eval", {"attrs": ["bold", "underline"]})

    # metrics
    metrics = OrderedDict(
        [
            ("operation_point", MetricApplyThresholds(pred="model.output.classification")),  # will apply argmax
            ("accuracy", MetricAccuracy(pred="results:metrics.operation_point.cls_pred", target="data.gt.probSevere")),
            (
                "roc",
                MetricROCCurve(
                    pred="model.output.classification",
                    target="data.gt.probSevere",
                    output_filename=os.path.join(paths["inference_dir"], "roc_curve.png"),
                ),
            ),
            ("auc", MetricAUCROC(pred="model.output.classification", target="data.gt.probSevere")),
        ]
    )

    # create evaluator
    evaluator = EvaluatorDefault()

    # run
    results = evaluator.eval(ids=None, data=infer_file, metrics=metrics, output_dir=paths["eval_dir"])

    return results


######################################
# Run
######################################
if __name__ == "__main__":
    # allocate gpus
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    GPU.choose_and_enable_multiple_gpus(NUM_GPUS, force_gpus=force_gpus)

    RUNNING_MODES = ["train", "infer", "eval"]  # Options: 'train', 'infer', 'eval'
    # train
    if "train" in RUNNING_MODES:
        run_train(paths=PATHS, train_common_params=TRAIN_COMMON_PARAMS)

    # infer
    if "infer" in RUNNING_MODES:
        run_infer(paths=PATHS, infer_common_params=INFER_COMMON_PARAMS)

    # eval
    if "eval" in RUNNING_MODES:
        run_eval(paths=PATHS, eval_common_params=EVAL_COMMON_PARAMS)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

==========================

MNIST classfier implementation that demonstrate end to end training, inference and evaluation using FuseMedML
This example shows how to directly train a model using custom (your own) pytorch lightning module implementation
"""
import copy
import logging
import os
from typing import Any, List, OrderedDict, Sequence, Tuple
from fuse.utils.ndict import NDict

import torch
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data.dataloader import DataLoader
import pytorch_lightning as pl

from fuse.eval.evaluator import EvaluatorDefault
from fuse.eval.metrics.classification.metrics_thresholding_common import MetricApplyThresholds
from fuse.eval.metrics.classification.metrics_classification_common import MetricAccuracy, MetricAUCROC, MetricROCCurve

from fuse.data.utils.samplers import BatchSamplerDefault
from fuse.data.utils.collates import CollateDefault

from fuse.dl.losses.loss_default import LossDefault
from fuse.dl.models.model_wrapper import ModelWrapSeqToDict
import fuse.dl.lightning.pl_funcs as fuse_pl

from fuse.utils.utils_debug import FuseDebug
from fuse.utils.utils_logger import fuse_logger_start
from fuse.utils.file_io.file_io import create_dir, save_dataframe
import fuse.utils.gpu as GPU

from fuseimg.datasets.mnist import MNIST

from fuse_examples.imaging.classification.mnist import lenet

###########################################################################################################
# Fuse
###########################################################################################################

##########################################
# Lightning Module
##########################################


class LightningModuleMnist(pl.LightningModule):
    """
    Implementation of pl.LightningModule
    Demonstrates how to use FuseMedML with your own PyTorch lightining implementaiton.
    """

    def __init__(self, model_dir: str, opt_lr: float, opt_weight_decay: float, **kwargs):
        """
        :param model_dir: location for checkpoints and logs
        :param opt_lr: learning rate for Adam optimizer
        :param opt_weight_decay: weight decay for Adam optimizer
        """
        super().__init__(**kwargs)
        self.save_hyperparameters(ignore=["model_dir"])

        # store arguments
        self._model_dir = model_dir
        self._opt_lr = opt_lr
        self._opt_weight_decay = opt_weight_decay

        # init state
        self._prediction_keys = None

        # model
        torch_model = lenet.LeNet()

        # wrap basic torch model to automatically read inputs from batch_dict and save its outputs to batch_dict
        self._model = ModelWrapSeqToDict(
            model=torch_model,
            model_inputs=["data.image"],
            post_forward_processing_function=perform_softmax,
            model_outputs=["model.logits.classification", "model.output.classification"],
        )

        # losses
        self._losses = {
            "cls_loss": LossDefault(
                pred="model.logits.classification", target="data.label", callable=F.cross_entropy, weight=1.0
            ),
        }

        # metrics
        self._train_metrics = OrderedDict(
            [
                ("operation_point", MetricApplyThresholds(pred="model.output.classification")),  # will apply argmax
                ("accuracy", MetricAccuracy(pred="results:metrics.operation_point.cls_pred", target="data.label")),
            ]
        )

        self._validation_metrics = copy.deepcopy(self._train_metrics)  # use the same metrics in validation as well

    ## forward
    def forward(self, batch_dict: NDict) -> NDict:
        return self._model(batch_dict)

    ## Step
    def training_step(self, batch_dict: NDict, batch_idx: int) -> dict:
        # run forward function and store the outputs in batch_dict["model"]
        batch_dict = self.forward(batch_dict)
        # given the batch_dict and FuseMedML style losses - compute the losses, return the total loss and save losses values in batch_dict["losses"]
        total_loss = fuse_pl.step_losses(self._losses, batch_dict)
        # given the batch_dict and FuseMedML style losses - collect the required values to compute the metrics on epoch_end
        fuse_pl.step_metrics(self._train_metrics, batch_dict)

        # return the total_loss, the losses and drop everything else
        return {"loss": total_loss, "losses": batch_dict["losses"]}

    def validation_step(self, batch_dict: NDict, batch_idx: int) -> dict:
        # run forward function and store the outputs in batch_dict["model"]
        batch_dict = self.forward(batch_dict)
        # given the batch_dict and FuseMedML style losses - compute the losses, return the total loss (ignored) and save losses values in batch_dict["losses"]
        _ = fuse_pl.step_losses(self._losses, batch_dict)
        # given the batch_dict and FuseMedML style losses - collect the required values to compute the metrics on epoch_end
        fuse_pl.step_metrics(self._validation_metrics, batch_dict)

        # return just the losses and drop everything else
        return {"losses": batch_dict["losses"]}

    def predict_step(self, batch_dict: NDict, batch_idx: int) -> dict:
        if self._prediction_keys is None:
            raise Exception(
                "Error: predict_step expectes list of prediction keys to extract from batch_dict. Please specify it using set_predictions_keys() method "
            )
        # run forward function and store the outputs in batch_dict["model"]
        batch_dict = self.forward(batch_dict)
        # extract the requried keys - defined in self.set_predictions_keys()
        return fuse_pl.step_extract_predictions(self._prediction_keys, batch_dict)

    ## Epoch end
    def training_epoch_end(self, step_outputs) -> None:
        # calc average epoch loss and log it
        fuse_pl.epoch_end_compute_and_log_losses(self, "train", [e["losses"] for e in step_outputs])
        # evaluate  and log it
        fuse_pl.epoch_end_compute_and_log_metrics(self, "train", self._train_metrics)

    def validation_epoch_end(self, step_outputs) -> None:
        # calc average epoch loss and log it
        fuse_pl.epoch_end_compute_and_log_losses(self, "validation", [e["losses"] for e in step_outputs])
        # evaluate  and log it
        fuse_pl.epoch_end_compute_and_log_metrics(self, "validation", self._validation_metrics)

    def configure_callbacks(self) -> Sequence[pl.Callback]:
        """Create callbacks to monitor the metrics and print epoch summary"""
        best_epoch_source = dict(
            monitor="validation.metrics.accuracy",
            mode="max",
        )
        return fuse_pl.model_checkpoint_callbacks(self._model_dir, best_epoch_source)

    def configure_optimizers(self) -> Any:
        """See pl.LightningModule.configure_optimizers return value for all options"""
        # create optimizer
        optimizer = optim.Adam(self._model.parameters(), lr=self._opt_lr, weight_decay=self._opt_weight_decay)

        # create learning scheduler
        lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer)
        lr_sch_config = dict(scheduler=lr_scheduler, monitor="validation.losses.total_loss")
        return dict(optimizer=optimizer, lr_scheduler=lr_sch_config)

    def set_predictions_keys(self, keys: List[str]) -> None:
        """Define which keys to extract from batch_dict  on prediction mode"""
        self._prediction_keys = keys


##########################################
# Debug modes
##########################################
mode = "default"  # Options: 'default', 'debug'. See details in FuseDebug
debug = FuseDebug(mode)

##########################################
# Output Paths
##########################################
ROOT = "_examples/mnist_custom"  # TODO: fill path here
model_dir = os.path.join(ROOT, "model_dir")
PATHS = {
    "model_dir": model_dir,
    "cache_dir": os.path.join(ROOT, "cache_dir"),
    "inference_dir": os.path.join(model_dir, "infer_dir"),
    "eval_dir": os.path.join(model_dir, "eval_dir"),
}

NUM_GPUS = 1
##########################################
# Train Common Params
##########################################
TRAIN_COMMON_PARAMS = {}

# ============
# Data
# ============
TRAIN_COMMON_PARAMS["data.batch_size"] = 100
TRAIN_COMMON_PARAMS["data.train_num_workers"] = 8
TRAIN_COMMON_PARAMS["data.validation_num_workers"] = 8

# ===============
# PL Trainer
# ===============
TRAIN_COMMON_PARAMS["trainer.num_epochs"] = 2
TRAIN_COMMON_PARAMS["trainer.num_devices"] = NUM_GPUS
TRAIN_COMMON_PARAMS["trainer.accelerator"] = "gpu"
# use "dp" strategy temp when working with multiple GPUS - workaround for pytorch lightning issue: https://github.com/Lightning-AI/lightning/issues/11807
TRAIN_COMMON_PARAMS["trainer.strategy"] = "dp" if TRAIN_COMMON_PARAMS["trainer.num_devices"] > 1 else None
TRAIN_COMMON_PARAMS["trainer.ckpt_path"] = None  # path to the checkpoint you wish continue the training from

# ===============
# PL Module
# ===============
TRAIN_COMMON_PARAMS["pl_module.opt_lr"] = 1e-4
TRAIN_COMMON_PARAMS["pl_module.opt_weight_decay"] = 0.001


def perform_softmax(logits: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    cls_preds = F.softmax(logits, dim=1)
    return logits, cls_preds


#################################
# Train Template
#################################
def run_train(paths: dict, train_params: dict):
    # ==============================================================================
    # Logger
    # ==============================================================================
    fuse_logger_start(output_path=paths["model_dir"], console_verbose_level=logging.INFO)
    print("Fuse Train")

    # ==============================================================================
    # Data
    # ==============================================================================
    # Train Data
    print("Data - trainset:")
    train_dataset = MNIST.dataset(paths["cache_dir"], train=True)
    print("- Create sampler:")
    sampler = BatchSamplerDefault(
        dataset=train_dataset,
        balanced_class_name="data.label",
        num_balanced_classes=10,
        batch_size=train_params["data.batch_size"],
        balanced_class_weights=None,
    )
    print("- Create sampler: Done")

    # Create dataloader
    train_dataloader = DataLoader(
        dataset=train_dataset,
        batch_sampler=sampler,
        collate_fn=CollateDefault(),
        num_workers=train_params["data.train_num_workers"],
    )
    print("Data - trainset: Done")

    ## Validation data
    print("Data - validation set:")
    # wrapping torch dataset
    validation_dataset = MNIST.dataset(paths["cache_dir"], train=False)

    # dataloader
    validation_dataloader = DataLoader(
        dataset=validation_dataset,
        batch_size=train_params["data.batch_size"],
        collate_fn=CollateDefault(),
        num_workers=train_params["data.validation_num_workers"],
    )
    print("Data - validation set: Done")

    # ==============================================================================
    # Train
    # ==============================================================================
    # create instance of PL module
    pl_module = LightningModuleMnist(
        model_dir=paths["model_dir"],
        opt_lr=train_params["pl_module.opt_lr"],
        opt_weight_decay=train_params["pl_module.opt_weight_decay"],
    )

    # create lightining trainer.
    pl_trainer = pl.Trainer(
        default_root_dir=paths["model_dir"],
        max_epochs=train_params["trainer.num_epochs"],
        accelerator=train_params["trainer.accelerator"],
        strategy=train_params["trainer.strategy"],
        devices=train_params["trainer.num_devices"],
        auto_select_gpus=True,
    )

    # train
    pl_trainer.fit(pl_module, train_dataloader, validation_dataloader, ckpt_path=train_params["trainer.ckpt_path"])
    print("Train: Done")


######################################
# Inference Common Params
######################################
INFER_COMMON_PARAMS = {}
INFER_COMMON_PARAMS["infer_filename"] = "infer_file.gz"
INFER_COMMON_PARAMS["checkpoint"] = "best_epoch.ckpt"
INFER_COMMON_PARAMS["trainer.num_devices"] = 1  # infer should use single device
INFER_COMMON_PARAMS["trainer.accelerator"] = "gpu"
INFER_COMMON_PARAMS["trainer.strategy"] = None

######################################
# Inference Template
######################################


def run_infer(paths: dict, infer_common_params: dict):
    create_dir(paths["inference_dir"])
    infer_file = os.path.join(paths["inference_dir"], infer_common_params["infer_filename"])
    checkpoint_file = os.path.join(paths["model_dir"], infer_common_params["checkpoint"])
    #### Logger
    fuse_logger_start(output_path=paths["model_dir"], console_verbose_level=logging.INFO)

    print("Fuse Inference")
    print(f"infer_filename={infer_file}")

    ## Data
    # Create dataset
    validation_dataset = MNIST.dataset(paths["cache_dir"], train=False)
    # dataloader
    validation_dataloader = DataLoader(
        dataset=validation_dataset, collate_fn=CollateDefault(), batch_size=2, num_workers=2
    )

    # load pytorch lightning module
    pl_module = LightningModuleMnist.load_from_checkpoint(
        checkpoint_file, model_dir=paths["model_dir"], map_location="cpu", strict=True
    )
    # set the prediction keys to extract (the ones used be the evaluation function).
    pl_module.set_predictions_keys(
        ["model.output.classification", "data.label"]
    )  # which keys to extract and dump into file

    print("Model: Done")
    # create a trainer instance
    pl_trainer = pl.Trainer(
        default_root_dir=paths["model_dir"],
        accelerator=infer_common_params["trainer.accelerator"],
        devices=infer_common_params["trainer.num_devices"],
        strategy=infer_common_params["trainer.strategy"],
        auto_select_gpus=True,
    )
    predictions = pl_trainer.predict(pl_module, validation_dataloader, return_predictions=True)

    # convert list of batch outputs into a dataframe
    infer_df = fuse_pl.convert_predictions_to_dataframe(predictions)
    save_dataframe(infer_df, infer_file)


######################################
# Analyze Common Params
######################################
EVAL_COMMON_PARAMS = {}
EVAL_COMMON_PARAMS["infer_filename"] = INFER_COMMON_PARAMS["infer_filename"]


######################################
# Eval Template
######################################
def run_eval(paths: dict, eval_common_params: dict):
    create_dir(paths["eval_dir"])
    infer_file = os.path.join(paths["inference_dir"], eval_common_params["infer_filename"])
    fuse_logger_start(output_path=None, console_verbose_level=logging.INFO)

    print("Fuse Eval")

    # metrics
    class_names = [str(i) for i in range(10)]

    metrics = OrderedDict(
        [
            ("operation_point", MetricApplyThresholds(pred="model.output.classification")),  # will apply argmax
            ("accuracy", MetricAccuracy(pred="results:metrics.operation_point.cls_pred", target="data.label")),
            (
                "roc",
                MetricROCCurve(
                    pred="model.output.classification",
                    target="data.label",
                    class_names=class_names,
                    output_filename=os.path.join(paths["eval_dir"], "roc_curve.png"),
                ),
            ),
            ("auc", MetricAUCROC(pred="model.output.classification", target="data.label", class_names=class_names)),
        ]
    )

    # create evaluator
    evaluator = EvaluatorDefault()

    # run
    results = evaluator.eval(ids=None, data=infer_file, metrics=metrics, output_dir=paths["eval_dir"])

    return results


######################################
# Run
######################################
if __name__ == "__main__":
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    GPU.choose_and_enable_multiple_gpus(NUM_GPUS, force_gpus=force_gpus)

    RUNNING_MODES = ["train", "infer", "eval"]  # Options: 'train', 'infer', 'eval'
    # train
    if "train" in RUNNING_MODES:
        run_train(paths=PATHS, train_params=TRAIN_COMMON_PARAMS)

    # infer
    if "infer" in RUNNING_MODES:
        run_infer(paths=PATHS, infer_common_params=INFER_COMMON_PARAMS)

    # eval
    if "eval" in RUNNING_MODES:
        run_eval(paths=PATHS, eval_common_params=EVAL_COMMON_PARAMS)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
import logging
import sys
import time
from IPython import get_ipython
from collections import Iterable
from threading import Lock

from typing import List, Union, Sequence, Optional, Hashable, Any

import numpy as np
import pandas as pd
from torch import Tensor

from fuse.utils.ndict import NDict


class Misc:
    @classmethod
    def flatten(cls, l: List[List]) -> List:
        """
        Flatten a list of lists, yielding a 1-d list of strings
        :param l: list of lists, possible irregular shapes
        :return: generator of single str elements
        """
        for el in l:
            if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
                yield from cls.flatten(el)
            else:
                yield el

    @staticmethod
    def query_yes_no(question: str, default: Optional[str] = None) -> bool:
        """
        Ask the user yes/no question
        :param question: string question
        :param default: default answer in case the user type enter. if set to none the question will be asked again.
        :return: the answer
        """
        # can't prompt for answer in notebooks, assume yes
        if Misc.in_ipynb():
            return True

        valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
        if default is None:
            prompt = " [y/n] "
        elif default == "yes":
            prompt = " [Y/n] "
        elif default == "no":
            prompt = " [y/N] "
        else:
            raise ValueError("invalid default answer: '%s'" % default)

        while True:
            sys.stdout.write(question + prompt)
            sys.stdout.flush()
            choice = sys.stdin.readline().rstrip("\r\n").lower()
            if default is not None and choice == "":
                return valid[default]
            elif choice in valid:
                return valid[choice]
            else:
                sys.stdout.write("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")

    @staticmethod
    def squeeze_obj(obj: Any) -> Any:
        """
        Get batch with single sample tensor / numpy / list and squeeze the batch dimension
        :param obj: the object to sqeeze
        :return: squeezed object
        """
        if isinstance(obj, Tensor):
            obj_tensor: Tensor = obj
            if obj_tensor.shape[0] == 1:
                obj_tensor = obj_tensor.squeeze(dim=0)
                return obj_tensor
        elif isinstance(obj, np.ndarray):
            obj_np: np.ndarray = obj
            if obj_np.shape[0] == 1:
                obj_np = obj_np.squeeze(axis=0)
                return obj_np
        elif isinstance(obj, list):
            obj_lst: list = obj
            if len(obj_lst) == 1:
                return obj_lst[0]

        return obj

    @staticmethod
    def batch_dict_to_string(batch_dict: NDict) -> str:
        """
        Convert batch dict to string, including the keys, types and shapes.
        :param batch_dict: might be any dict
        :return: string representation of the batch dict
        """
        res = ""
        all_keys = batch_dict.keypaths()
        for key in all_keys:
            value = batch_dict[key]
            res += f"{key} : type={type(value)}"
            if isinstance(value, (np.ndarray, Tensor)):
                res += f", dtype={value.dtype}, shape={value.shape}"
            elif isinstance(value, Sequence):
                res += f", length={len(value)}"
            res += "\n"
        return res

    @staticmethod
    def in_ipynb():
        try:
            ip = get_ipython()
            if ip is None:
                return False
            else:
                return True
        except NameError:
            return False


def time_display(seconds, granularity=3):
    intervals = (
        # ('weeks', 604800),  # 60 * 60 * 24 * 7
        ("days", 86400),  # 60 * 60 * 24
        ("hours", 3600),  # 60 * 60
        ("minutes", 60),
        ("seconds", 1),
    )
    if seconds < 1:
        return "< second"
    result = []
    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            if value == 1:
                name = name.rstrip("s")
            result.append("%d %s" % (value, name))
    return ", ".join(result[:granularity])


def get_pretty_dataframe(df, col_width=25):
    # check is col_width needs to be widen (so that dashes are in one line)
    max_val_width = np.vectorize(len)(df.values.astype(str)).max()  # get maximum length of all values
    max_col_width = max([len(x) for x in df.columns])  # get the maximum lengths of all column names
    col_width = max(max_col_width, max_val_width, col_width)

    dashes = (col_width + 2) * len(df.columns.values)
    df_as_string = f"\n{'-' * dashes}\n"
    for col in df.columns.values:
        df_as_string += f"| {col:{col_width}}"
    df_as_string += f"|\n{'-' * dashes}\n"
    for idx, row in df.iterrows():
        for col in df.columns.values:
            df_as_string += f"| {row[col]:{col_width}}"

        df_as_string += f"|\n{'-' * dashes}\n"
    return df_as_string


def get_time_delta(begin_time: float) -> str:
    """
    Returns a string representing the difference in time between now and the begin_time

    :param begin_time: start
    :return: diff between now and begin_time
    """
    time_as_float = time.time() - begin_time
    return time_display(time_as_float)


def autodetect_input_source(input_source: Union[str, pd.DataFrame, Sequence[Hashable]] = None):
    """
    Loads sample descriptors from multiple auto-detectable possible sources:
    1. DataFrame (instance or path to pickled object)
    2. Python list of sample descriptors
    3. Text file (needs to end with '.txt' or '.text' extension)
    :param input_source:
    :return:
    """

    df = None

    # DataFrame instance
    # ------------------
    if isinstance(input_source, pd.DataFrame):
        df = input_source

    # Python list of sample descriptors
    # ---------------------------------
    elif isinstance(input_source, (list, tuple)):
        df = pd.DataFrame({"sample_desc": input_source})

    # Path to:
    # --------
    elif isinstance(input_source, str):
        file_extension = input_source.lower().split(".")[-1]

        # Text file
        # ---------
        if file_extension in ["text", "txt"]:
            with open(input_source, "r") as f:
                lines = f.readlines()
                sample_descs = [line.strip() for line in lines]
                df = pd.DataFrame({"sample_desc": sample_descs})

        # Pickled DataFrame
        # -----------------
        if file_extension in ["p", "pkl", "pickle", "gz", "bz2", "zip", "xz"]:
            df = pd.read_pickle(input_source)

    return df


class Singleton(type):
    """
    Singleton metaclass
    """

    _instances = {}
    _lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                cls._instances[cls] = super().__call__(*args, **kwargs)
            else:
                if len(args) + len(kwargs) > 0:
                    logging.getLogger("Fuse").warning(
                        "Ignoring a redefinition of the singleton of class {class_.__name__}"
                    )

            return cls._instances[cls]



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import math
from typing import Any, List, Optional, Union

import numpy as np
from torch.utils.data.sampler import Sampler

from fuse.data.datasets.dataset_base import DatasetBase


class BatchSamplerDefault(Sampler):
    """
    Torch batch sampler - balancing per batch
    """

    def __init__(
        self,
        dataset: DatasetBase,
        balanced_class_name: str,
        num_balanced_classes: int,
        batch_size: Optional[int] = None,
        mode: str = "exact",
        balanced_class_weights: Union[List[int], List[float], None] = None,
        num_batches: Optional[int] = None,
        **dataset_get_multi_kwargs,
    ) -> None:
        """
        :param dataset: dataset used to extract the balanced class from each sample
        :param balanced_class_name:  the name of balanced class to extract from dataset
        :param num_balanced_classes: number of classes to balance between
        :param batch_size: batch_size.
                        - In "exact" mode
                            If balanced_class_weights is None, must be set and divided by num_balanced_classes. Otherwise keep None.
                        - In "approx" mode
                          Must be set
        :param mode: either 'exact' or 'approx'. if 'exact each element in balanced_class_weights will specify the exact number of samples from this class.
                        if 'approx' - each element will specify the a probability that a sample will be from this class
        :param balanced_class_weights: Optional, integer/float per balanced class, Expected length is num_balanced_classes.
                                        In mode 'exact' expecting list of integers that sums up to batch dict.
                                        In mode 'approx' expecting list of floats that sums up to ~1
                                        If not specified and equal number of samples from each class will be used.
        :param num_batches: optional - if set will force num_batches, otherwise num_batches will be automatically to go over each sample at least once (exactly or approximately).
        :param dataset_get_multi_kwargs: extra parameters for dataset.get_multi() to optimize the running time.
        """
        super().__init__(None)

        # store input
        self._mode = mode
        self._dataset = dataset
        self._balanced_class_name = balanced_class_name
        self._num_balanced_classes = num_balanced_classes
        self._batch_size = batch_size
        self._balanced_class_weights = balanced_class_weights
        self._num_batches = num_batches
        self._dataset_get_multi_kwargs = dataset_get_multi_kwargs
        # modify relevant keys
        if self._balanced_class_name not in self._dataset_get_multi_kwargs:
            self._dataset_get_multi_kwargs["keys"] = [self._balanced_class_name]

        # validate input
        # modes
        if self._mode not in ["exact", "approx"]:
            raise Exception("Error, expected sampler mode to be either 'exact' or 'approx', got {mode}")

        # weights
        if self._mode == "exact":
            if self._balanced_class_weights is not None:
                for weight in self._balanced_class_weights:
                    if not isinstance(weight, int):
                        raise Exception(
                            f"Error: in mode 'exact', expecting only integers in balanced_class_weights, got {type(weight)}"
                        )
                if self._batch_size is not None:
                    if self._batch_size != sum(self._balanced_class_weights):
                        raise Exception(
                            f"Error: in mode 'exact', expecting balanced_class_weights {self._balanced_class_weights} to sum up to batch size {self._batch_size}. Consider setting batch_size to None to automatically compute the batch size."
                        )
                else:
                    self._batch_size = sum(self._balanced_class_weights)
            elif self._batch_size is None:
                raise Exception("Error: In 'approx' mode, either batch_size or balanced_class_weights")

        if self._mode == "approx":
            if self._batch_size is None:
                raise Exception("Error: in mode 'approx', batch size must be set.")
            if balanced_class_weights is not None:
                for weight in balanced_class_weights:
                    if not isinstance(weight, float):
                        raise Exception(
                            f"Error: in mode 'exact', expecting only floats in balanced_class_weights, got {type(weight)}"
                        )
                if not math.isclose(sum(self._balanced_class_weights), 1.0):
                    raise Exception(
                        f"Error: in mode 'exact', expecting balanced_class_weight to sum up to almost one, got {balanced_class_weights}"
                    )

        if balanced_class_weights is not None:
            if len(balanced_class_weights) != num_balanced_classes:
                raise Exception(
                    f"Expecting balance_class_weights ({balanced_class_weights}) to have a weight per balanced class ({num_balanced_classes})"
                )
        if self._mode == "exact":
            if self._batch_size < self._num_balanced_classes:
                raise Exception(
                    f"Error: batch_size ({self._batch_size}) should be greater than or equal to num_balanced_class ({self._num_balanced_classes})."
                )

        # if weights not specified, set weights to equally balance per batch
        if self._balanced_class_weights is None:
            if self._mode == "exact":
                if self._batch_size % self._num_balanced_classes:
                    raise Exception(
                        f"Error: num_balanced_class ({self._num_balanced_classes}) should devide batch_size ({self._batch_size}) in exact mode."
                    )

                self._balanced_class_weights = [
                    self._batch_size // self._num_balanced_classes
                ] * self._num_balanced_classes
            elif self._mode == "approx":
                self._balanced_class_weights = [1 / self._num_balanced_classes] * self._num_balanced_classes

        # get balanced classes per each sample
        collected_data = dataset.get_multi(None, **self._dataset_get_multi_kwargs)
        self._balanced_classes = self._extract_balanced_classes(collected_data)

        # split samples to groups
        self._balanced_class_indices = [
            np.where(self._balanced_classes == cls_i)[0] for cls_i in range(self._num_balanced_classes)
        ]
        self._balanced_class_sizes = [
            len(self._balanced_class_indices[cls_i]) for cls_i in range(self._num_balanced_classes)
        ]

        # make sure that size != 0 for all balanced classes
        for cls_ind, cls_size in enumerate(self._balanced_class_sizes):
            if self._balanced_class_weights[cls_ind] != 0.0 and cls_size == 0:
                msg = f"Every balanced class must include at least one sample (num of samples per balanced class{self._balanced_class_sizes} and weights are {self._balanced_class_weights})"
                raise Exception(msg)

        # calc batch index to balanced class mapping according to weights
        if self._mode == "exact":
            self._batch_index_to_class = []
            for balanced_cls in range(self._num_balanced_classes):
                self._batch_index_to_class.extend([balanced_cls] * self._balanced_class_weights[balanced_cls])
        else:
            # probabilistic method - will be randomly select per epoch
            self._batch_index_to_class = None

        # Shuffle balanced class indices
        for indices in self._balanced_class_indices:
            np.random.shuffle(indices)

        # Calculate num batches. Number of batches to iterate over all data at least once (exactly or approximately)
        # Calculate only if not directly specified by the user
        if self._num_batches is None:
            if self._mode == "exact":
                samples_per_batch = self._balanced_class_weights
            else:  # mode is approx
                # approximate size!
                samples_per_batch = [val * self._batch_size for val in self._balanced_class_weights]
            balanced_class_weighted_sizes = [
                math.ceil(self._balanced_class_sizes[cls_i] / samples_per_batch[cls_i])
                if self._balanced_class_weights[cls_i] != 0
                else 0
                for cls_i in range(self._num_balanced_classes)
            ]
            bigger_balanced_class_weighted_size = max(balanced_class_weighted_sizes)
            self._num_batches = int(bigger_balanced_class_weighted_size)

        # pointers per class
        self._cls_pointers = [0] * self._num_balanced_classes
        self._sample_pointer = 0

    def __iter__(self) -> np.ndarray:
        for batch_idx in range(self._num_batches):
            yield self._make_batch()

    def __len__(self) -> int:
        return self._num_batches

    def _get_sample(self, balanced_class: int) -> Any:
        """
        sample index given balanced class value
        :param balanced_class: integer representing balanced class value
        :return: sample index
        """
        sample_idx = self._balanced_class_indices[balanced_class][self._cls_pointers[balanced_class]]

        self._cls_pointers[balanced_class] += 1
        if self._cls_pointers[balanced_class] == self._balanced_class_sizes[balanced_class]:
            self._cls_pointers[balanced_class] = 0
            np.random.shuffle(self._balanced_class_indices[balanced_class])

        return sample_idx

    def _make_batch(self) -> list:
        """
        :return: list of indices to collate batch
        """
        if self._mode == "exact":
            batch_index_to_class = self._batch_index_to_class
        else:  # mode == approx
            # calc one according to probabilities
            batch_index_to_class = np.random.choice(
                np.arange(self._num_balanced_classes), self._batch_size, p=self._balanced_class_weights
            )
        batch_sample_indices = []
        for batch_index in range(self._batch_size):
            balanced_class = batch_index_to_class[batch_index]
            batch_sample_indices.append(self._get_sample(balanced_class))

        np.random.shuffle(batch_sample_indices)
        return batch_sample_indices

    def _extract_balanced_classes(self, collected_data: List[dict]) -> np.ndarray:
        """
        Extracting balanced class values from collected data.
        If - special extra logic is required. Either override this method or the logic in Op and append to dataset pipeline
        """
        assert len(collected_data) > 0, "Error: sampling failed, dataset size is 0"
        balanced_classes = [sample[self._balanced_class_name] for sample in collected_data]
        return np.array(balanced_classes)



"""
(C) Copyright 2021 IBM Corp.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
   http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Created on June 30, 2021
"""

from typing import Any, Callable, Optional, List, Sequence, Union

import numpy
import torch
import warnings

warnings.warn("FuseUtilsHierarchicalDict is obsolete and soon will be deleted! Please consider trasform to NDict.")


class FuseUtilsHierarchicalDict:
    @classmethod
    def get(cls, hierarchical_dict: dict, key: str):
        """
        get(dict, 'x.y.z') <==> dict['x']['y']['z']
        """
        # split according to '.'
        hierarchical_key = key.split(".")

        # go over the the dictionary towards the requested value
        try:
            value = hierarchical_dict[hierarchical_key[0]]
            for sub_key in hierarchical_key[1:]:
                value = value[sub_key]
            return value
        except:
            flat_dict = FuseUtilsHierarchicalDict.flatten(hierarchical_dict)
            if key in flat_dict:
                return flat_dict[key]
            else:
                raise KeyError(f"key {key} does not exist\n. Possible keys are: {str(list(flat_dict.keys()))}")

    @classmethod
    def set(cls, hierarchical_dict: dict, key: str, value: Any) -> None:
        """
        set(dict, 'x.y.z', value) <==> dict['x']['y']['z'] = value
        If either 'x', 'y' or 'z' nodes do not exist, this function will create them
        """
        # split according to '.'
        hierarchical_key = key.split(".")

        # go over the the dictionary according to the path, create the nodes that does not exist
        element = hierarchical_dict
        for key in hierarchical_key[:-1]:
            if key not in element:
                element[key] = {}
            element = element[key]

        # set the value
        element[hierarchical_key[-1]] = value

    @classmethod
    def get_all_keys(cls, hierarchical_dict: dict, include_values: bool = False) -> Union[List[str], dict]:
        """
        Get all hierarchical keys in  hierarchical_dict
        """
        all_keys = {}
        for key in hierarchical_dict:
            if isinstance(hierarchical_dict[key], dict):
                all_sub_keys = FuseUtilsHierarchicalDict.get_all_keys(hierarchical_dict[key], include_values=True)
                keys_to_add = {f"{key}.{sub_key}": all_sub_keys[sub_key] for sub_key in all_sub_keys}
                all_keys.update(keys_to_add)
            else:
                all_keys[key] = hierarchical_dict[key]
        if include_values:
            return all_keys
        else:
            return list(all_keys.keys())

    @classmethod
    def subkey(cls, key: str, start: int, end: Optional[int]) -> Optional[str]:
        """
        Sub string of hierarchical key.
        Example: subkey('a.b.c.d.f', 1, 3) -> 'b.c'
        :param key: the original key
        :param start: start index
        :param end: end index, not including
        :return: str
        """
        key_parts = key.split(".")

        # if end not specified set to max.
        if end is None:
            end = len(key_parts)

        if len(key_parts) < start or len(key_parts) < end:
            return None

        res = ".".join(key_parts[start:end])
        return res

    @classmethod
    def apply_on_all(cls, hierarchical_dict: dict, apply_func: Callable, *args: Any) -> None:
        all_keys = cls.get_all_keys(hierarchical_dict)
        for key in all_keys:
            new_value = apply_func(cls.get(hierarchical_dict, key), *args)
            cls.set(hierarchical_dict, key, new_value)
        pass

    @classmethod
    def flatten(cls, hierarchical_dict: dict) -> dict:
        """
        Flatten the dict
        @param hierarchical_dict: dict to flatten
        @return: dict where keys are the hierarchical_dict keys separated by periods.
        """
        return cls.get_all_keys(hierarchical_dict, include_values=True)

    @classmethod
    def indices(cls, hierarchical_dict: dict, indices: List[int]) -> dict:
        """
        Extract the specified indices from each element in the dictionary (if possible)
        :param hierarchical_dict: input dict
        :param indices: indices to extract
        :return: dict with only the required indices
        """
        new_dict = {}
        all_keys = cls.get_all_keys(hierarchical_dict)
        for key in all_keys:
            value = cls.get(hierarchical_dict, key)
            if isinstance(value, numpy.ndarray) or isinstance(value, torch.Tensor):
                new_value = value[indices]
            elif isinstance(value, Sequence):
                new_value = [item for i, item in enumerate(value) if indices[i]]
            else:
                new_value = value
            cls.set(new_dict, key, new_value)
        return new_dict

    @classmethod
    def to_string(cls, hierarchical_dict: dict) -> str:
        """
        Get flat string including thr content of the dictionary
        :param hierarchical_dict: input dict
        :return: string
        """
        keys = cls.get_all_keys(hierarchical_dict)
        keys = sorted(keys)
        res = ""
        for key in keys:
            res += f"{key} = {FuseUtilsHierarchicalDict.get(hierarchical_dict, key)}\n"

        return res

    @classmethod
    def pop(cls, hierarchical_dict: dict, key: str):
        """
        return the value hierarchical_dict[key] and remove the key from the dict.
        :param hierarchical_dict: the dictionary
        :param key: the key to return and remove
        """
        # split according to '.'
        hierarchical_key = key.split(".")
        # go over the the dictionary towards the requested value
        try:
            key_idx = len(hierarchical_key) - 1
            value = hierarchical_dict[hierarchical_key[0]] if key_idx > 0 else hierarchical_dict
            for sub_key in hierarchical_key[1:-1]:
                value = value[sub_key]
            return value.pop(hierarchical_key[key_idx])
        except:
            flat_dict = FuseUtilsHierarchicalDict.flatten(hierarchical_dict)
            if key in flat_dict:
                return flat_dict[key]
            else:
                raise KeyError(f"key {key} does not exist\n. Possible keys are: {str(list(flat_dict.keys()))}")

    @classmethod
    def is_in(cls, hierarchical_dict: dict, key: str) -> bool:
        """
        Returns True if the full key is in dict, False otherwise.
        e.g., for dict = {'a':1, 'b.c':2} is_in(dict, 'b.c') returns True, but is_in(dict, 'c') returns False.
        :param hierarchical_dict: dict to check
        :param key: key to search
        :return: key in hierarchical_dict
        """
        return key in cls.get_all_keys(hierarchical_dict)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import unittest

import pandas as pd

from fuse.utils.misc.misc import get_pretty_dataframe


class PrettyDataframeTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_normal_strings(self):
        df = pd.DataFrame({"A": ["abc", "de", "efg"], "B": ["a", "abcde", "abc"], "C": [1, 2.5, 1.5]})

        col_width = 25
        df_as_string = get_pretty_dataframe(df)

        self.assertTrue("abcd" in df_as_string)
        self.assertTrue(f'|{" " * (col_width - 2)}1.0|\n' in df_as_string)
        self.assertTrue(f'| abcde{" " * (col_width - 5)}|' in df_as_string)
        self.assertTrue(f"|\n{'-' * (col_width + 2) * 3}\n" in df_as_string)

    def test_long_strings(self):
        df = pd.DataFrame(
            {
                "A": ["abc", "de", "very long column name should be longer than 25"],
                "B": ["a", "abcde", "abc"],
                "C": [1, 2.5, 1.5],
            }
        )

        col_width = len("very long column name should be longer than 25")
        df_as_string = get_pretty_dataframe(df)

        self.assertTrue("abcd" in df_as_string)
        self.assertTrue(f'|{" " * (col_width - 2)}1.0|\n' in df_as_string)
        self.assertTrue(f'| abcde{" " * (col_width - 5)}|' in df_as_string)
        self.assertTrue(f"|\n{'-' * (col_width + 2) * 3}\n" in df_as_string)

    def test_long_headers(self):
        df = pd.DataFrame(
            {
                "very long column name should be longer than 25": ["abc", "de", "abcd"],
                "B": ["a", "abcde", "abc"],
                "C": [1, 2.5, 1.5],
            }
        )

        col_width = len("very long column name should be longer than 25")
        df_as_string = get_pretty_dataframe(df)

        self.assertTrue("abcd" in df_as_string)
        self.assertTrue(f'|{" " * (col_width - 2)}1.0|\n' in df_as_string)
        self.assertTrue(f'| abcde{" " * (col_width - 5)}|' in df_as_string)
        self.assertTrue(f"|\n{'-' * (col_width + 2) * 3}\n" in df_as_string)

    def tearDown(self):
        pass


if __name__ == "__main__":
    unittest.main()



import os
from glob import glob
from typing import Hashable, Optional, Sequence, Tuple

from fuse.utils.rand.param_sampler import Uniform, RandInt, RandBool
from fuse.utils.ndict import NDict

from fuse.data import DatasetDefault
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuse.data import PipelineDefault, OpToTensor
from fuse.data.ops.op_base import OpBase
from fuse.data.ops.ops_aug_common import OpSample, OpRandApply
from fuse.data.ops.ops_common import OpLambda, OpZScoreNorm
from fuseimg.data.ops.aug.geometry import OpAugAffine2D, OpRotation3D, OpResizeTo
from fuseimg.data.ops.aug.color import OpAugGaussian
from fuseimg.data.ops.image_loader import OpLoadImage
from fuseimg.data.ops.color import OpClip
import numpy as np
from fuse.data.utils.sample import get_sample_id
from functools import partial
import torch
import pandas as pd


class OpKnightSampleIDDecode(OpBase):
    """
    decodes sample id into image and segmentation filename
    """

    def __call__(self, sample_dict: NDict, test: bool = False) -> NDict:
        """ """

        sid = get_sample_id(sample_dict)
        sample_dict["data.input.case_id"] = sid
        img_filename_key = "data.input.img_path"
        if test:
            sample_dict[img_filename_key] = f"images/{sid}.nii.gz"
        else:
            sample_dict[img_filename_key] = os.path.join(sid, "imaging.nii.gz")

            seg_filename_key = "data.gt.seg_path"
            sample_dict[seg_filename_key] = os.path.join(sid, "aggregated_MAJ_seg.nii.gz")

        return sample_dict


class OpClinicalLoad(OpBase):
    def __init__(self, json_path: str):
        super().__init__()
        self.json_path = json_path

    def __call__(self, sample_dict: NDict, test: bool = False) -> NDict:
        cols = [
            "case_id",
            "age_at_nephrectomy",
            "body_mass_index",
            "gender",
            "comorbidities",
            "smoking_history",
            "radiographic_size",
            "last_preop_egfr",
            "voxel_spacing",
        ]

        if test:
            json_data = pd.read_json(os.path.join(self.json_path, "features.json"))[cols]
        else:
            cols += ["aua_risk_group"]
            json_data = pd.read_json(os.path.join(self.json_path, "knight.json"))[cols]

        sid = sample_dict["data.input.case_id"]
        row = json_data[json_data["case_id"] == sid].to_dict("records")[0]

        row["gender"] = int(row["gender"].lower() == "female")  # female:1 | male:0
        row["comorbidities"] = int(
            any(x for x in row["comorbidities"].values())
        )  # if has any comorbidity it is set to 1
        row["smoking_history"] = ["never_smoked", "previous_smoker", "current_smoker"].index(row["smoking_history"])
        if row["last_preop_egfr"] is None or row["last_preop_egfr"]["value"] is None:
            row["last_preop_egfr"] = 77  # median value
        elif row["last_preop_egfr"]["value"] in (">=90", ">90"):
            row["last_preop_egfr"] = 90
        else:
            row["last_preop_egfr"] = row["last_preop_egfr"]["value"]

        if row["radiographic_size"] is None:
            row["radiographic_size"] = 4.1  # this is the median value on the training set
        if not test:
            sample_dict["data.gt.gt_global.task_1_label"] = int(
                row["aua_risk_group"] in ["high_risk", "very_high_risk"]
            )
            sample_dict["data.gt.gt_global.task_2_label"] = [
                "benign",
                "low_risk",
                "intermediate_risk",
                "high_risk",
                "very_high_risk",
            ].index(row["aua_risk_group"])
        row["voxel_spacing"] = (
            row["voxel_spacing"]["z_spacing"],
            row["voxel_spacing"]["y_spacing"],
            row["voxel_spacing"]["x_spacing"],
        )

        sample_dict["data.input.clinical"] = row
        return sample_dict


class OpPrepareClinical(OpBase):
    def __call__(
        self, sample_dict: NDict
    ) -> NDict:  # , op_id: Optional[str]) -> NDict:, op_id: Optional[str]) -> NDict:
        age = sample_dict["data.input.clinical.age_at_nephrectomy"]
        if age is not None and age > 0 and age < 120:
            age = np.array(age / 120.0).reshape(-1)
        else:
            age = np.array(-1.0).reshape(-1)

        bmi = sample_dict["data.input.clinical.body_mass_index"]
        if bmi is not None and bmi > 10 and bmi < 100:
            bmi = np.array(bmi / 50.0).reshape(-1)
        else:
            bmi = np.array(-1.0).reshape(-1)

        radiographic_size = sample_dict["data.input.clinical.radiographic_size"]
        if radiographic_size is not None and radiographic_size > 0 and radiographic_size < 50:
            radiographic_size = np.array(radiographic_size / 15.0).reshape(-1)
        else:
            radiographic_size = np.array(-1.0).reshape(-1)

        preop_egfr = sample_dict["data.input.clinical.last_preop_egfr"]
        if preop_egfr is not None and preop_egfr > 0 and preop_egfr < 200:
            preop_egfr = np.array(preop_egfr / 90.0).reshape(-1)
        else:
            preop_egfr = np.array(-1.0).reshape(-1)
        # turn categorical features into one hot vectors
        gender = sample_dict["data.input.clinical.gender"]
        gender_one_hot = np.zeros(len(GENDER_INDEX))
        if gender in GENDER_INDEX.values():
            gender_one_hot[gender] = 1

        comorbidities = sample_dict["data.input.clinical.comorbidities"]
        comorbidities_one_hot = np.zeros(len(COMORBIDITIES_INDEX))
        if comorbidities in COMORBIDITIES_INDEX.values():
            comorbidities_one_hot[comorbidities] = 1

        smoking_history = sample_dict["data.input.clinical.smoking_history"]
        smoking_history_one_hot = np.zeros(len(SMOKE_HISTORY_INDEX))
        if smoking_history in SMOKE_HISTORY_INDEX.values():
            smoking_history_one_hot[smoking_history] = 1

        clinical_encoding = np.concatenate(
            (age, bmi, radiographic_size, preop_egfr, gender_one_hot, comorbidities_one_hot, smoking_history_one_hot),
            axis=0,
            dtype=np.float32,
        )
        sample_dict["data.input.clinical.all"] = clinical_encoding
        return sample_dict


class KNIGHT:
    """
    Dataset created for KNIGHT challenge - https://research.ibm.com/haifa/Workshops/KNIGHT/challenge.html
    Aims to predict the risk level of patients based on CT scan and/or clinical data.
    """

    @staticmethod
    def sample_ids(path: str) -> list:
        """
        get all the sample ids in train-set
        sample_id is directory file named case_xxxxx found in the specified path
        """
        files = [f for f in glob(os.path.join(path, "data/case_*"))]
        return files

    @staticmethod
    def static_pipeline(data_path: str, resize_to: Tuple, test: bool = False) -> PipelineDefault:
        static_pipeline = PipelineDefault(
            "static",
            [
                # decoding sample ID
                (
                    OpKnightSampleIDDecode(),
                    dict(test=test),
                ),  # will save image and seg path to "data.input.img_path", "data.gt.seg_path" and load json data
                (OpClinicalLoad(data_path), dict(test=test)),
                # loading data
                (OpLoadImage(data_path), dict(key_in="data.input.img_path", key_out="data.input.img", format="nib")),
                # fixed image normalization
                (OpClip(), dict(key="data.input.img", clip=(-62, 301))),
                (OpZScoreNorm(), dict(key="data.input.img", mean=104.0, std=75.3)),  # kits normalization
                # transposing so the depth channel will be first
                (
                    OpLambda(partial(np.moveaxis, source=-1, destination=0)),
                    dict(key="data.input.img"),
                ),  # convert image from shape [H, W, D] to shape [D, H, W]
                (OpPrepareClinical(), dict()),  # process clinical data
                (OpResizeTo(channels_first=False), dict(key="data.input.img", output_shape=resize_to)),
            ],
        )
        return static_pipeline

    @staticmethod
    def train_dynamic_pipeline() -> PipelineDefault:
        train_dynamic_pipeline = PipelineDefault(
            "dynamic",
            [
                # Numpy to tensor
                (OpToTensor(), dict(key="data.input.img", dtype=torch.float)),
                (OpToTensor(), dict(key="data.input.clinical.all")),
                (
                    OpRandApply(OpSample(OpRotation3D()), 0.5),
                    dict(
                        key="data.input.img",
                        z_rot=Uniform(-5.0, 5.0),
                        x_rot=Uniform(-5.0, 5.0),
                        y_rot=Uniform(-5.0, 5.0),
                    ),
                ),
                # affine transformation per slice but with the same arguments
                (
                    OpRandApply(OpSample(OpAugAffine2D()), 0.5),
                    dict(
                        key="data.input.img",
                        rotate=Uniform(-180.0, 180.0),
                        scale=Uniform(0.8, 1.2),
                        flip=(RandBool(0.5), RandBool(0.5)),
                        translate=(RandInt(-15, 15), RandInt(-15, 15)),
                    ),
                ),
                (OpAugGaussian(), dict(key="data.input.img", std=0.01)),
                # add channel dimension -> [C=1, D, H, W]
                (OpLambda(partial(torch.unsqueeze, dim=0)), dict(key="data.input.img")),
            ],
        )
        return train_dynamic_pipeline

    @staticmethod
    def val_dynamic_pipeline() -> PipelineDefault:
        val_dynamic_pipeline = PipelineDefault(
            "dynamic",
            [
                # Numpy to tensor
                (OpToTensor(), dict(key="data.input.img", dtype=torch.float)),
                (OpToTensor(), dict(key="data.input.clinical.all")),
                # add channel dimension -> [C=1, D, H, W]
                (OpLambda(partial(torch.unsqueeze, dim=0)), dict(key="data.input.img")),
            ],
        )
        return val_dynamic_pipeline

    @staticmethod
    def dataset(
        data_path: str = "data",
        cache_dir: str = "cache",
        split: dict = None,
        sample_ids: Optional[Sequence[Hashable]] = None,
        test: bool = False,
        reset_cache: bool = False,
        resize_to: Tuple = (70, 256, 256),
    ) -> DatasetDefault:
        """
        Get cached dataset
        :param data_path: path to store the original data
        :param cache_dir: path to store the cache
        :param split: dictionary including sample ids for (train and validation) or test.
        :param sample_ids: dataset including the specified sample_ids. sample_id is case_{id:05d} (for example case_00001 or case_00100).
        either split or sample_ids is not None. there is no need in both of them.
        :param test: boolean indicating weather to use train dynamic pipeline or val. only necessary when using sample_ids param.
        :param reset_cache: set to True tp reset the cache
        :param train: True if used for training  - adds augmentation operations to the pipeline
        """
        train_dynamic_pipeline = KNIGHT.train_dynamic_pipeline()
        val_dynamic_pipeline = KNIGHT.val_dynamic_pipeline()

        # Create dataset
        if sample_ids is not None:
            static_pipeline = KNIGHT.static_pipeline(data_path, resize_to=resize_to, test=test)
            cacher = SamplesCacher(
                "cache", static_pipeline, cache_dirs=[f"{cache_dir}/data"], restart_cache=reset_cache, workers=8
            )
            dataset = DatasetDefault(
                sample_ids=sample_ids,
                static_pipeline=static_pipeline,
                dynamic_pipeline=val_dynamic_pipeline if test else train_dynamic_pipeline,
                cacher=cacher,
            )
            print("- Load and cache data:")
            dataset.create()
            print("- Load and cache data: Done")
            return dataset

        static_pipeline = KNIGHT.static_pipeline(data_path, resize_to=resize_to, test=("test" in split))
        if "train" in split:
            train_cacher = SamplesCacher(
                "train_cache", static_pipeline, cache_dirs=[f"{cache_dir}/train"], restart_cache=reset_cache, workers=8
            )

            train_dataset = DatasetDefault(
                sample_ids=split["train"],
                static_pipeline=static_pipeline,
                dynamic_pipeline=train_dynamic_pipeline,
                cacher=train_cacher,
            )

            print("- Load and cache data:")
            train_dataset.create()

            print("- Load and cache data: Done")

            print("Train Data: Done", {"attrs": "bold"})

            #### Validation data
            print("Validation Data:", {"attrs": "bold"})

            val_cacher = SamplesCacher(
                "val_cache", static_pipeline, cache_dirs=[f"{cache_dir}/val"], restart_cache=reset_cache, workers=8
            )
            ## Create dataset
            validation_dataset = DatasetDefault(
                sample_ids=split["val"],
                static_pipeline=static_pipeline,
                dynamic_pipeline=val_dynamic_pipeline,
                cacher=val_cacher,
            )

            print("- Load and cache data:")
            validation_dataset.create()
            print("- Load and cache data: Done")

            print("Validation Data: Done", {"attrs": "bold"})

            return train_dataset, validation_dataset
        else:  # test only
            #### Test data
            print("Test Data:", {"attrs": "bold"})

            ## Create dataset
            test_dataset = DatasetDefault(
                sample_ids=split["test"],
                static_pipeline=static_pipeline,
                dynamic_pipeline=val_dynamic_pipeline,
            )

            print("- Load and cache data:")
            test_dataset.create()
            print("- Load and cache data: Done")

            print("Test Data: Done", {"attrs": "bold"})
            return test_dataset


GENDER_INDEX = {"male": 0, "female": 1}
COMORBIDITIES_INDEX = {"no comorbidities": 0, "comorbidities exist": 1}
SMOKE_HISTORY_INDEX = {"never smoked": 0, "previous smoker": 1, "current smoker": 2}



import runpy
from typing import List, Union, Callable, Dict


class Config:
    def __init__(self):
        pass

    def load(self, *configs: List[Union[Dict, str, Callable]]) -> Dict:
        """
        loads, in order, configs
        each item can be either:
        1. Python dict. the python dict "accumulated" so for will be dict.update-ed with this element dict
        2. path to a python script file. It will be expected to contain a callable:
            def get_config(config:dict) -> dict
                ...
        3. a callable with the signature (config:dict) -> dict
        """
        ans = dict()
        for idx, conf in enumerate(configs):
            if isinstance(conf, dict):
                ans.update(conf)
            elif isinstance(conf, str):
                func = get_config_function(conf)
                ans = func(ans)
            elif callable(conf):
                ans = conf(ans)
            else:
                raise Exception(f"config element #{idx} is not a valid type. It has type={type(conf)}")
        return ans


def get_config_function(script_path: str):
    ans = runpy.run_path(script_path)
    func_name = "load_config"
    if func_name not in ans:
        raise Exception(f"Expected to have load_config(conf:dict) -> dict defined in {script_path}")
    return ans[func_name]



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
import pathlib
import sys
import os
from typing import List, Tuple, Union, Optional
from collections import OrderedDict

import csv
import pandas as pd
import numpy as np

from fuse.utils.ndict import NDict
from fuse.utils.file_io.file_io import create_dir

from fuse.eval.evaluator import EvaluatorDefault

from fuse.eval.metrics.classification.metrics_classification_common import (
    MetricAUCROC,
    MetricROCCurve,
)
from fuse.eval.metrics.metrics_common import CI


## Constants
# Constants that defines the expected format of the prediction and target files and list the classes for task 1 and task @
EXPECTED_TASK1_PRED_KEYS = {"case_id", "NoAT-score", "CanAT-score"}
EXPECTED_TASK2_PRED_KEYS = {"case_id", "B-score", "LR-score", "IR-score", "HR-score", "VHR-score"}
EXPECTED_TARGET_KEYS = {"case_id", "Task1-target", "Task2-target"}
PRED_CASE_ID_NAME = "case_id"
TARGET_CASE_ID_NAME = "case_id"
TASK1_CLASS_NAMES = ("NoAT", "CanAT")  # must be aligned with task1 targets
TASK2_CLASS_NAMES = ("B", "LR", "IR", "HR", "VHR")  # must be aligned with task2 targets


def post_processing(sample_dict: NDict, task1: bool = True, task2: bool = True) -> dict:
    """
    post caching processing. Will group together to an array the per class scores and verify it sums up to 1.0
    :param sample_dict: a dictionary that contais all the extracted values of a single sample
    :param task1: if true will evaluate task1
    :param task2: if true will evaluate task2
    :return: a modified/alternative dictionary
    """
    # verify sample
    expected_keys = [f"target.{key}" for key in EXPECTED_TARGET_KEYS]
    if task1:
        expected_keys += [f"task1_pred.{key}" for key in EXPECTED_TASK1_PRED_KEYS]
    if task2:
        expected_keys += [f"task2_pred.{key}" for key in EXPECTED_TASK2_PRED_KEYS]
    set(expected_keys).issubset(set(sample_dict.keypaths()))

    # convert scores to numpy array
    # task 1
    if task1:
        task1_pred = []
        for cls_name in TASK1_CLASS_NAMES:
            task1_pred.append(sample_dict[f"task1_pred.{cls_name}-score"])
        task1_pred_array = np.array(task1_pred)
        if not np.isclose(task1_pred_array.sum(), 1.0, rtol=0.05):
            print(
                f"Warning: expecting task 1 prediction for case {sample_dict['descriptor']} to sum up to almost 1.0, got {task1_pred_array}"
            )
        sample_dict["task1_pred.array"] = task1_pred_array

    # task 2
    if task2:
        task2_pred = []
        for cls_name in TASK2_CLASS_NAMES:
            task2_pred.append(sample_dict[f"task2_pred.{cls_name}-score"])
        task2_pred_array = np.array(task2_pred)
        if not np.isclose(task2_pred_array.sum(), 1.0, rtol=0.05):
            print(
                f"Error: expecting task 2 prediction for case {sample_dict['descriptor']} to sum up to almost 1.0, got {task2_pred_array}"
            )
        sample_dict["task2_pred.array"] = task2_pred_array

    return sample_dict


def decode_results(results: NDict, output_dir: str, task1: bool, task2: bool) -> Tuple[OrderedDict, str]:
    """
    Gets the results computed by the dictionary and summarize it in a markdown text and dictionary.
    The dictionary will be saved in <output_dir>/results.csv and the markdown text in <output_dir>/results.md
    :param results: the results computed by the metrics
    :param output_dir: path to an output directory
    :param task1: if true will evaluate task1
    :param task2: if true will evaluate task2
    :return: ordered dict summarizing the results and markdown text
    """
    results = NDict(results["metrics"])
    results_table = OrderedDict()
    # Table
    ## task1
    if task1:
        results_table["Task1-AUC"] = f"{results['task1_auc.macro_avg.org']:.3f}"
        results_table[
            "Task1-AUC-CI"
        ] = f"[{results['task1_auc.macro_avg.conf_lower']:.3f}-{results['task1_auc.macro_avg.conf_upper']:.3f}]"

    ## task2
    if task2:
        results_table["Task2-AUC"] = f"{results['task2_auc.macro_avg.org']:.3f}"
        results_table[
            "Task2-AUC-CI"
        ] = f"[{results['task2_auc.macro_avg.conf_lower']:.3f}-{results['task2_auc.macro_avg.conf_upper']:.3f}]"
        for cls_name in TASK2_CLASS_NAMES:
            results_table[f"Task2-AUC-{cls_name}VsRest"] = f"{results[f'task2_auc.{cls_name}.org']:.3f}"
            results_table[
                f"Task2-AUC-{cls_name}VsRest-CI"
            ] = f"[{results[f'task2_auc.{cls_name}.conf_lower']:.3f}-{results[f'task2_auc.{cls_name}.conf_upper']:.3f}]"

    # mark down text
    results_text = ""
    ## task 1
    if task1:
        results_text += "# Task 1 - adjuvant treatment candidacy classification\n"
        results_text += f"AUC: {results_table['Task1-AUC']} {results_table['Task1-AUC-CI']}\n"
        results_text += "## ROC Curve\n"
        results_text += '<br/>\n<img src="task1_roc.png" alt="drawing" width="40%"/>\n<br/>\n'

    ## task 2
    if task2:
        results_text += "\n# Task 2 - risk categories classification\n"
        results_text += f"AUC: {results_table['Task2-AUC']} {results_table['Task2-AUC-CI']}\n"

        results_text += "## Multi-Class AUC\n"
        table_columns = ["AUC"] + [f"AUC-{cls_name}VsRest" for cls_name in TASK2_CLASS_NAMES]
        results_text += "\n|"
        results_text += "".join([f" {column} |" for column in table_columns])
        results_text += "\n|"
        results_text += "".join([" ------ |" for column in table_columns])
        results_text += "\n|"
        results_text += "".join(
            [f" {results_table[f'Task2-{column}']} {results_table[f'Task2-{column}-CI']} |" for column in table_columns]
        )
        results_text += "\n## ROC Curve\n"
        results_text += '<br/>\n<img src="task2_roc.png" alt="drawing" width="40%"/>\n<br/>\n'

    # save files
    with open(os.path.join(output_dir, "results.md"), "w") as output_file:
        output_file.write(results_text)

    with open(os.path.join(output_dir, "results.csv"), "w") as output_file:
        w = csv.writer(output_file)
        w.writerow(results_table.keys())
        w.writerow(results_table.values())

    return results_table, results_text


def eval(
    task1_prediction_filename: str,
    task2_prediction_filename: str,
    target_filename: str,
    output_dir: str,
    case_ids_source: Optional[Union[str, List[str]]] = "target",
) -> Tuple[OrderedDict, str]:
    """
    Load the prediction and target files, evaluate the predictions and save the results into files.
    :param task1_prediction_filename: path to a prediction csv file for task1. Expecting the columns listed in EXPECTED_TASK1_PRED_KEYS to exist (including the header).
                                      if set to "" - the script will not evaluate task1
    :param task2_prediction_filename: path to a prediction csv file for task2. Expecting the columns listed in EXPECTED_TASK2_PRED_KEYS to exist (including the header)
                                      if set to "" - the script will not evaluate task2
    :param target_filename: path to a prediction csv file. Expecting the columns listed in TARGET_CASE_ID_NAME to exist
    :param output_dir: path to directory to save the output files
    :param case_ids_source: will run the evaluation on the specified list of case ids. Supported values:
                                 * "task1_pred" to evaluate all the samples/cases specified in task1 prediction file.
                                 * "task2_pred" to evaluate all the samples/cases specified in task2 prediction file.
                                 * "target" to evaluate all the samples/cases specified in targets file.
                                 * list to define the samples explicitly
    :return: ordered dict summarizing the results and markdown text
    """
    create_dir(output_dir)

    # eval task1, task2 or both
    task1 = task1_prediction_filename is not None and task1_prediction_filename != ""
    task2 = task2_prediction_filename is not None and task2_prediction_filename != ""

    if case_ids_source is None:
        if task1:
            case_ids_source = "task1_pred"
        else:
            case_ids_source = "task2_pred"

    dataframes_dict = {}
    metrics = {}
    # task 1
    if task1:
        # metrics to evaluate
        metrics.update(
            {
                "task1_auc": CI(
                    MetricAUCROC(
                        pred="task1_pred.array",
                        target="target.Task1-target",
                        class_names=TASK1_CLASS_NAMES,
                        pre_collect_process_func=post_processing,
                    ),
                    stratum="target.Task1-target",
                ),
                "task1_roc_curve": MetricROCCurve(
                    pred="task1_pred.array",
                    target="target.Task1-target",
                    class_names=[None, ""],
                    pre_collect_process_func=post_processing,
                    output_filename=os.path.join(output_dir, "task1_roc.png"),
                ),
            }
        )
        # read files
        task1_pred_df = pd.read_csv(task1_prediction_filename, dtype={PRED_CASE_ID_NAME: object})
        # verify input
        assert set(task1_pred_df.keys()).issubset(
            EXPECTED_TASK1_PRED_KEYS
        ), f"Expecting task1 prediction file {os.path.abspath(task1_prediction_filename)} to include also the following keys: {EXPECTED_TASK1_PRED_KEYS - set(task1_pred_df.keys())}"
        task1_pred_df["id"] = task1_pred_df[PRED_CASE_ID_NAME]
        dataframes_dict["task1_pred"] = task1_pred_df

    # task 2
    if task2:
        # metrics to evaluate
        metrics.update(
            {
                "task2_auc": CI(
                    MetricAUCROC(
                        pred="task2_pred.array",
                        target="target.Task2-target",
                        class_names=TASK2_CLASS_NAMES,
                        pre_collect_process_func=post_processing,
                    ),
                    stratum="target.Task2-target",
                ),
                "task2_roc_curve": MetricROCCurve(
                    pred="task2_pred.array",
                    target="target.Task2-target",
                    class_names=TASK2_CLASS_NAMES,
                    output_filename=os.path.join(output_dir, "task2_roc.png"),
                    pre_collect_process_func=post_processing,
                ),
            }
        )
        # read files
        task2_pred_df = pd.read_csv(task2_prediction_filename, dtype={PRED_CASE_ID_NAME: object})
        # verify input
        assert set(task2_pred_df.keys()).issubset(
            EXPECTED_TASK2_PRED_KEYS
        ), f"Expecting task2 prediction file {os.path.abspath(task2_prediction_filename)} to include also the following keys: {EXPECTED_TASK2_PRED_KEYS - set(task2_pred_df.keys())}"
        task2_pred_df["id"] = task2_pred_df[PRED_CASE_ID_NAME]
        dataframes_dict["task2_pred"] = task2_pred_df

    # read files
    target_df = pd.read_csv(target_filename, dtype={TARGET_CASE_ID_NAME: object})
    # verify input
    assert set(target_df.keys()).issubset(
        EXPECTED_TARGET_KEYS
    ), f"Expecting target file {os.path.abspath(target_filename)} to include also the following keys: {EXPECTED_TARGET_KEYS - set(target_df.keys())}"
    target_df["id"] = target_df[TARGET_CASE_ID_NAME]
    dataframes_dict["target"] = target_df

    # analyze
    evaluator = EvaluatorDefault()
    results = evaluator.eval(
        ids=list(dataframes_dict[case_ids_source]["id"]), data=dataframes_dict, metrics=metrics, output_dir=None
    )

    # output
    return decode_results(results, output_dir=output_dir, task1=task1, task2=task2)


if __name__ == "__main__":
    """
    Run evaluation:
    Usage: python eval.py <target_filename> <task1 prediction_filename> <task2 prediction_filename> <output dir>
    See details in function eval()
    Run dummy example (set the working dir to fuse-med-ml/examples/fuse_examples/imaging/classification/knight/eval): python eval.py example/example_targets.csv example/example_task1_predictions.csv example/example_task2_predictions.csv example/results
    """
    if len(sys.argv) == 1:
        dir_path = pathlib.Path(__file__).parent.resolve()
        # no arguments - set arguments inline - see details in function eval()
        target_filename = os.path.join(dir_path, "example/example_targets.csv")
        task1_prediction_filename = os.path.join(dir_path, "example/example_task1_predictions.csv")
        task2_prediction_filename = os.path.join(dir_path, "example/example_task2_predictions.csv")
        output_dir = "example/result"
    else:
        # get arguments from sys.argv
        assert (
            len(sys.argv) == 5
        ), f"Error: expecting 4 input arguments, but got {len(sys.argv)-1}. Usage: python eval.py <target_filename> <task1_prediction_filename> <task2_prediction_filename> <output_dir>. See details in function eval()"
        target_filename = sys.argv[1]
        task1_prediction_filename = sys.argv[2]
        task2_prediction_filename = sys.argv[3]
        output_dir = sys.argv[4]

    eval(
        target_filename=target_filename,
        task1_prediction_filename=task1_prediction_filename,
        task2_prediction_filename=task2_prediction_filename,
        output_dir=output_dir,
    )



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from typing import List, Tuple, Union, Optional
from fuse.data.ops.op_base import OpBase, OpReversibleBase, op_call, op_reverse
from fuse.utils.misc.context import DummyContext
from fuse.utils.ndict import NDict
from fuse.utils.cpu_profiling.timer import Timer


class PipelineDefault(OpReversibleBase):
    """
    Pipeline default implementation
    Pipeline to run sequence of ops with a dictionary passing information between the ops.
    See OpBase for more information
    """

    def __init__(
        self,
        name: str,
        ops_and_kwargs: List[Tuple[OpBase, dict]],
        op_ids: Optional[List[str]] = None,
        verbose: bool = False,
    ):
        """
        :param name: pipeline name
        :param ops_and_args: List of tuples. Each tuple include op and dictionary includes op specific arguments.
        :param op_ids: Optional, set op_id - unique name for every op. If not set, an index will be used
        :param verbose: set to True for debug messages such as the running time of each operation
        """
        super().__init__()
        self._name = name
        self._ops_and_kwargs = ops_and_kwargs
        if op_ids is None:
            self._op_ids = [str(index) for index in range(len(self._ops_and_kwargs))]
        else:
            assert len(self._ops_and_kwargs) == len(op_ids), "Expecting op_id for every op"
            assert len(set(op_ids)) == len(op_ids), "Expecting unique op id for every op."
            self._op_ids = op_ids
        self._verbose = verbose

    def extend(self, ops_and_kwargs: List[Tuple[OpBase, dict]], op_ids: Optional[List[str]] = None):
        """
        Extends pipeline
        :param ops_and_args: Ops to append, List of tuples. Each tuple include op and dictionary includes op specific arguments.
        :param op_ids: Optional, set op_id - unique name for every op. If not set, an index will be used
        """
        if op_ids is None:
            op_ids = [str(index + len(self._ops_and_kwargs)) for index in range(len(ops_and_kwargs))]
        else:
            assert len(ops_and_kwargs) == len(op_ids), "Expecting op_id for every op"
            all_op_ids = self._op_ids + op_ids
            assert len(set(all_op_ids)) == len(all_op_ids), "Expecting unique op id for every op."

        self._ops_and_kwargs.extend(ops_and_kwargs)
        self._op_ids.extend(op_ids)

    def get_name(self) -> str:
        return self._name

    def __str__(self) -> str:
        text = []
        for (op_id, op_kwargs) in zip(self._op_ids, self._ops_and_kwargs):
            op, kwargs = op_kwargs
            text.append(str(op_id) + "@" + op.get_hashable_string_representation() + "@" + str(kwargs) + "@")
        return "".join(text)  # this is faster than accumulate_str+=new_str

    def __call__(
        self, sample_dict: NDict, op_id: Optional[str] = None, until_op_id: Optional[str] = None
    ) -> Union[None, dict, List[dict]]:
        """
        See super class
        plus
        :param until_op_id: optional - stop after the specified op_id - might be used for optimization
        """
        # set op_id if not specified
        if op_id is None:
            op_id = f"internal.{self._name}"

        samples_to_process = [sample_dict]
        for sub_op_id, (op, op_kwargs) in zip(self._op_ids, self._ops_and_kwargs):
            if self._verbose:
                context = Timer(f"Pipeline {self._name}: op {type(op).__name__}, op_id {sub_op_id}", self._verbose)
            else:
                context = DummyContext()
            with context:
                samples_to_process_next = []

                for sample in samples_to_process:

                    sample = op_call(op, sample, f"{op_id}.{sub_op_id}", **op_kwargs)

                    # three options for return value:
                    # None - ignore the sample
                    # List of dicts - split sample
                    # dict - modified sample
                    if sample is None:
                        return None
                    elif isinstance(sample, list):
                        samples_to_process_next += sample
                    elif isinstance(sample, dict):
                        samples_to_process_next.append(sample)
                    else:
                        raise Exception(f"unexpected sample type returned by {type(op)}: {type(sample)}")

            # continue to process with next op
            samples_to_process = samples_to_process_next

            # if required - stop after the specified op id
            if until_op_id is not None and sub_op_id == until_op_id:
                break

        # if single sample - return it, otherwise return list of samples.
        if len(samples_to_process) == 1:
            return samples_to_process[0]
        else:
            return samples_to_process

    def reverse(self, sample_dict: NDict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str] = None) -> dict:
        """
        See super class
        """
        # set op_id if not specified
        if op_id is None:
            op_id = f"internal.{self._name}"

        for sub_op_id, (op, _) in zip(reversed(self._op_ids), reversed(self._ops_and_kwargs)):
            sample_dict = op_reverse(op, sample_dict, f"{op_id}.{sub_op_id}", key_to_reverse, key_to_follow)

        return sample_dict



from .file_io import (
    save_pickle,
    load_pickle,
    save_pickle_safe,
    save_text_file_safe,
    save_text_file,
    read_simple_float_file,
    read_simple_int_file,
    read_text_file,
    get_randomized_postfix_name,
    read_single_str_line_file,
    create_simple_timestamp_file,
    save_hdf5_safe,
    load_hdf5,
    delete_directory_tree,
)

from .compressed import extract_zip_file

from .path import change_extension, get_extension, remove_extension



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import os
from typing import OrderedDict
import logging

import torch
import pytorch_lightning as pl
import torch.optim as optim
from torch.utils.data.dataloader import DataLoader

from fuse.utils.utils_debug import FuseDebug
import fuse.utils.gpu as GPU
from fuse.utils.utils_logger import fuse_logger_start
from fuse.utils.file_io.file_io import create_dir, save_dataframe

from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuse.data.datasets.dataset_default import DatasetDefault
from fuse.data.pipelines.pipeline_default import PipelineDefault
from fuse.data.utils.collates import CollateDefault
from fuse.data.utils.samplers import BatchSamplerDefault

from fuse.dl.models import ModelMultiHead
from fuse.dl.lightning.pl_module import LightningModuleDefault
from fuse.dl.lightning.pl_funcs import convert_predictions_to_dataframe

from fuse.eval.evaluator import EvaluatorDefault

###########################################################################################################
# Fuse
# Example of a training template:
#
# Training template contains 5 fundamental building blocks:
# (1) Data - detailed README file can be found at [fuse/data](../../fuse/data)
# (2) Model
# (3) Losses
# (4) Metrics and Evaluation - detailed README file can be found at [fuse/eval](../../fuse/eval)
# (5) Train - using PyTorch Lightning
#
# The template will create each of those building blocks and will eventually run pl_trainer.fit()
#
# Terminology:
# 1. sample-id -
#    A unique identifier of a sample. Each sample in the dataset must have an id that uniquely identifies it.
#    Examples of sample ids:
#    * path to the image file
#    * Tuple of (provider_id, patient_id, image_id)
#    * Running index
# 2. sample_dict -
#    Represents a single sample and contains all relevant information about the sample.
#    No specific structure of this dictionary is required, but a useful pattern is to split it into sections (keys that define a "namespace" ): such as "data", "model",  etc.
#    NDict (fuse/utils/ndict.py) class is used instead of python standard dictionary in order to allow easy "." separated access. For example:
#    `sample_dict[“data.input.img”]` is the equivalent of `sample_dict["data"]["input"]["img"]`
#    Another recommended convention is to include suffix specifying the type of the value ("img", "seg", "bbox")
# 3. Fuse base classes - *Base -
#    Abstract classes of the object forming together Fuse framework .
# 4. Fuse default classes - *Default -
#    A default generic implementation of the equivalent base class.
#    Those generic implementation will be useful for most common use cases.
#    Alternative implementations could be implemented for the special cases.
###########################################################################################################

##########################################
# Debug modes
##########################################
mode = "default"  # Options: 'default', 'debug'. See details in FuseDebug
debug = FuseDebug(mode)

##########################################
# Output Paths
##########################################
ROOT = None  # TODO: fill in a path to model dir
model_dir = os.path.join(ROOT, "model_dir")
PATHS = {
    "model_dir": model_dir,
    "force_reset_model_dir": False,  # If True will reset model dir automatically - otherwise will prompt 'are you sure' message.
    "cache_dir": os.path.join(ROOT, "cache_dir"),
    "inference_dir": os.path.join(model_dir, "infer"),
    "eval_dir": os.path.join(model_dir, "eval"),
    "data_split_filename": os.path.join(ROOT, "split.pkl"),
}

##########################################
# Train Common Params
##########################################
TRAIN_COMMON_PARAMS = {}
# ============
# Data
# ============
TRAIN_COMMON_PARAMS["data.batch_size"] = 2
TRAIN_COMMON_PARAMS["data.train_num_workers"] = 8
TRAIN_COMMON_PARAMS["data.validation_num_workers"] = 8
TRAIN_COMMON_PARAMS["data.cache_num_workers"] = 10

# ===============
# PL Trainer
# ===============
TRAIN_COMMON_PARAMS["trainer.num_epochs"] = 100
TRAIN_COMMON_PARAMS["trainer.num_devices"] = 1
TRAIN_COMMON_PARAMS["trainer.accelerator"] = "gpu"
TRAIN_COMMON_PARAMS["trainer.ckpt_path"] = None

# ===============
# Optimizer
# ===============
TRAIN_COMMON_PARAMS["opt.lr"] = 1e-4
TRAIN_COMMON_PARAMS["opt.weight_decay"] = 1e-3

# ===================================================================================================================
# Model
#   Build a model (torch.nn.Module) using generic Fuse components:
#   1. ModelMultiHead - generic component supporting single backbone with multiple heads
#   2. Backbone - simple backbone model
#   3. Head* - generic head implementations
#   The model outputs will be aggregated in batch_dict['model.*']
#   Each head output will be aggregated in batch_dict['model.<head name>.*']
#
#   Additional implemented models:
#   * ModelWrapper - allow to use single standard PyTorch module as is - see (examples/fuse_examples/imaging/classification/mnist/runner.py)[../../examples/fuse_examples/imaging/classification/mnist/runner.py]
#   * ModelEnsemble - runs several sub-modules sequentially
#   * ModelMultistream - convolutional neural network with multiple processing streams and multiple heads
# ===================================================================================================================


def create_model() -> torch.nn.Module:
    # TODO - define / create a model
    model = ModelMultiHead(
        conv_inputs=(("data.input.input_0.tensor", 1),),
        backbone="TODO",  # Reference: BackboneInceptionResnetV2
        heads=["TODO"],  # References: HeadGlobalPoolingClassifier, HeadDenseSegmentation
    )
    return model


#################################
# Train Template
#################################
def run_train(paths: dict, train_common_params: dict) -> None:
    # ==============================================================================
    # Logger
    #   - output log automatically to three destinations:
    #     (1) console (2) file - copy of the console (3) verbose file - used for debug
    #   - save a copy of the template file
    # ==============================================================================
    fuse_logger_start(output_path=paths["model_dir"], console_verbose_level=logging.INFO)

    print("Fuse Train")
    print(f'model_dir={paths["model_dir"]}')
    print(f'cache_dir={paths["cache_dir"]}')

    # ==============================================================================
    # Data
    #   Build dataloaders (torch.utils.data.DataLoader) for both train and validation.
    #
    #   Default dataset implementation built from a sequence of op(erator)s.
    #   Operators are the building blocks of the sample processing pipeline.
    #   Each operator gets as input the *sample_dict* as created by the previous operators and can either add/delete/modify fields in sample_dict.
    #   The operator interface is specified in OpBase class.
    #
    #   We split the pipeline into two parts - static and dynamic, which allow us to control the part out of the entire pipeline that will be cached.
    #
    #   For more details and examples, read (fuse/data/README.md)[../../fuse/data/README.md]
    #   A complete example implementation of a dataset can be bound in  (fuseimg/datasets/stoic21.py)[../../examples/fuse_examples/imaging/classification/stoic21/runner_stoic21.py] STOIC21.static_pipeline().
    # ==============================================================================

    #### Train Data

    print("Train Data:")

    ## TODO - list your sample ids:
    # Fuse TIP - splitting the sample_ids to folds can be done by fuse.data.utils.split.dataset_balanced_division_to_folds().
    #            See (examples/fuse_examples/imaging/classification/stoic21/runner_stoic21.py)[../../examples/fuse_examples/imaging/classification/stoic21/runner_stoic21.py]
    train_sample_ids = None
    validation_sample_ids = None

    ## Create data static_pipeline -
    #                                the output of this pipeline will be cached to optimize the running time and to better utilize the GPU:
    #                                See example in (fuseimg/datasets/stoic21.py)[../../examples/fuse_examples/imaging/classification/stoic21/runner_stoic21.py] STOIC21.static_pipeline().
    static_pipeline = PipelineDefault(
        "template_static",
        [
            # TODO
        ],
    )
    ## Create data dynamic_pipeline - Dynamic pipeline follows the static pipeline and continues to pre-process the sample.
    #                                 In contrast to the static pipeline, the output of the dynamic pipeline is not be cached and allows modifying the pre-precessing steps without recaching,
    #                                 The recommendation is to include in the dynamic pipeline pre-processing steps that we intend to experiment with or augmentation steps.
    train_dynamic_pipeline = PipelineDefault(
        "template_dynamic",
        [
            # TODO
        ],
    )
    validation_dynamic_pipeline = PipelineDefault(
        "template_dynamic",
        [
            # TODO
        ],
    )

    # Create dataset
    cacher = SamplesCacher(
        "template_cache",
        static_pipeline,
        [paths["cache_dir"]],
        restart_cache=False,
        workers=train_common_params["data.cache_num_workers"],
    )

    train_dataset = DatasetDefault(
        sample_ids=train_sample_ids,
        static_pipeline=static_pipeline,
        dynamic_pipeline=train_dynamic_pipeline,
        cacher=cacher,
    )

    print("- Load and cache data:")
    train_dataset.create()
    print("- Load and cache data: Done")

    ## Create batch sampler
    # Fuse TIPs:
    # 1. You don't have to balance according the classification labels, any categorical value will do.
    #    Use balanced_class_name to select the categorical value
    # 2. You don't have to equally balance between the classes.
    #    Use balanced_class_weights to specify the number of required samples in a batch per each class
    # 3. Use mode to specify probabilities rather then exact number of samples from  a class in each batch
    print("- Create sampler:")
    sampler = BatchSamplerDefault(
        dataset=train_dataset,
        balanced_class_name="TODO",
        num_balanced_classes="TODO",
        batch_size=train_common_params["data.batch_size"],
        balanced_class_weights=None,
    )

    print("- Create sampler: Done")

    ## Create dataloader
    train_dataloader = DataLoader(
        dataset=train_dataset,
        shuffle=False,
        drop_last=False,
        batch_sampler=sampler,
        collate_fn=CollateDefault(),
        num_workers=train_common_params["data.train_num_workers"],
    )
    print("Train Data: Done")

    #### Validation data
    print("Validation Data:")

    validation_dataset = DatasetDefault(
        sample_ids=validation_sample_ids,
        static_pipeline=static_pipeline,
        dynamic_pipeline=validation_dynamic_pipeline,
        cacher=cacher,
    )

    print("- Load and cache data:")
    validation_dataset.create()
    print("- Load and cache data: Done")

    ## Create dataloader
    validation_dataloader = DataLoader(
        dataset=validation_dataset,
        shuffle=False,
        drop_last=False,
        batch_sampler=None,
        batch_size=train_common_params["data.batch_size"],
        num_workers=train_common_params["data.validation_num_workers"],
        collate_fn=CollateDefault(),
    )
    print("Validation Data: Done")

    ## Create model
    print("Model:")
    model = create_model()
    print("Model: Done")

    # ==========================================================================================================================================
    #   Loss
    #   Dictionary of loss elements each element is a sub-class of LossBase
    #   The total loss will be the weighted sum of all the elements.
    #   Each element output loss will be aggregated in batch_dict['losses.<loss name>']
    #   The average batch loss per epoch will be included in epoch_result['losses.<loss name>'],
    #   and the total loss in epoch_result['losses.total_loss']
    #   The 'best_epoch_source', used to save the best model could be based on one of this losses.
    #   Available Losses:
    #   LossDefault - wraps a PyTorch loss function with an api.
    #
    # ==========================================================================================================================================
    losses = {
        # TODO add losses here (instances of LossBase)
    }

    # =========================================================================================================
    # Metrics - details can be found in (fuse/eval/README.md)[../../fuse/eval/README.md]
    #   1. Create seperately for train and validation (might be a deep copy, but not a shallow one).
    #   2. Set best_epoch_source:
    #       monitor: the metric name to track
    #       mode: either consider the "min" value to be best or the "max" value to be the best
    # =========================================================================================================
    train_metrics = OrderedDict(
        [
            # TODO add metrics here (<name>, <instance of MetricBase>)
        ]
    )
    validation_metrics = OrderedDict(
        [
            # TODO add metrics here (<name>, <instance of MetricBase>)
        ]
    )

    best_epoch_source = dict(monitor="TODO", mode="TODO")

    # =====================================================================================
    #  Train - using PyTorch Lightning
    #  Create training objects, PL module and PL trainer.
    # =====================================================================================
    print("Fuse Train:")

    # create optimizer
    optimizer = optim.Adam(
        model.parameters(),
        lr=train_common_params["opt.lr"],
        weight_decay=train_common_params["opt.weight_decay"],
    )

    # create scheduler
    lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer)
    lr_sch_config = dict(scheduler=lr_scheduler, monitor="validation.losses.total_loss")

    # optimizier and lr sch - see pl.LightningModule.configure_optimizers return value for all options
    optimizers_and_lr_schs = dict(optimizer=optimizer, lr_scheduler=lr_sch_config)

    # create instance of PL module - FuseMedML generic version
    pl_module = LightningModuleDefault(
        model_dir=paths["model_dir"],
        model=model,
        losses=losses,
        train_metrics=train_metrics,
        validation_metrics=validation_metrics,
        best_epoch_source=best_epoch_source,
        optimizers_and_lr_schs=optimizers_and_lr_schs,
    )

    # create lightining trainer.
    pl_trainer = pl.Trainer(
        default_root_dir=paths["model_dir"],
        max_epochs=train_common_params["trainer.num_epochs"],
        accelerator=train_common_params["trainer.accelerator"],
        devices=train_common_params["trainer.num_devices"],
        auto_select_gpus=True,
    )

    # train
    pl_trainer.fit(
        pl_module, train_dataloader, validation_dataloader, ckpt_path=train_common_params["trainer.ckpt_path"]
    )

    print("Fuse Train: Done")


######################################
# Inference Common Params
######################################
INFER_COMMON_PARAMS = {}
INFER_COMMON_PARAMS["data.num_workers"] = TRAIN_COMMON_PARAMS["data.train_num_workers"]
INFER_COMMON_PARAMS["data.batch_size"] = 4
INFER_COMMON_PARAMS["infer_filename"] = os.path.join(PATHS["inference_dir"], "validation_set_infer.pickle")
INFER_COMMON_PARAMS["checkpoint"] = "best_epoch.ckpt"  # Fuse TIP: possible values are 'best', 'last' or epoch_index.

######################################
# Inference Template
######################################


def run_infer(paths: dict, infer_common_params: dict) -> None:
    create_dir(paths["inference_dir"])
    infer_file = os.path.join(paths["inference_dir"], infer_common_params["infer_filename"])
    checkpoint_file = os.path.join(paths["model_dir"], infer_common_params["checkpoint"])

    ## Logger
    fuse_logger_start(output_path=paths["inference_dir"], console_verbose_level=logging.INFO)
    print("Fuse Inference")
    print(f"infer_filename={infer_file}")

    #### create infer dataset
    infer_dataset = None  # TODO: follow the same steps to create dataset as in run_train

    ## Create dataloader
    infer_dataloader = DataLoader(
        dataset=infer_dataset,
        shuffle=False,
        drop_last=False,
        batch_sampler=None,
        batch_size=infer_common_params["data.batch_size"],
        num_workers=infer_common_params["data.num_workers"],
        collate_fn=CollateDefault(),
    )

    # TODO - define / create a model
    model = ModelMultiHead(
        conv_inputs=(("data.input.input_0.tensor", 1),),
        backbone="TODO",  # Reference: BackboneInceptionResnetV2
        heads=["TODO"],  # References: HeadGlobalPoolingClassifier, HeadDenseSegmentation
    )

    # load python lightning module
    pl_module = LightningModuleDefault.load_from_checkpoint(
        checkpoint_file, model_dir=paths["model_dir"], model=model, map_location="cpu", strict=True
    )

    # set the prediction keys to extract and dump into file (the ones used be the evaluation function).
    pl_module.set_predictions_keys(
        [
            # TODO
        ]
    )

    # create a trainer instance and predict
    pl_trainer = pl.Trainer(
        default_root_dir=paths["model_dir"],
        accelerator=infer_common_params["trainer.accelerator"],
        devices=infer_common_params["trainer.num_devices"],
        auto_select_gpus=True,
    )
    predictions = pl_trainer.predict(pl_module, infer_dataloader, return_predictions=True)

    # convert list of batch outputs into a dataframe
    infer_df = convert_predictions_to_dataframe(predictions)
    save_dataframe(infer_df, infer_file)

    print("Fuse Inference: Done")


######################################
# Eval Template
######################################
EVAL_COMMON_PARAMS = {}
EVAL_COMMON_PARAMS["infer_filename"] = INFER_COMMON_PARAMS["infer_filename"]


def run_eval(paths: dict, eval_common_params: dict) -> None:
    fuse_logger_start(output_path=None, console_verbose_level=logging.INFO)
    print("Fuse Eval")

    # metrics
    metrics = OrderedDict(
        [
            # TODO add metrics here (<name>, <instance of MetricBase>)
        ]
    )

    # create evaluator
    evaluator = EvaluatorDefault()

    # run
    results = evaluator.eval(
        ids=None,
        data=os.path.join(paths["inference_dir"], eval_common_params["infer_filename"]),
        metrics=metrics,
        output_dir=paths["eval_dir"],
    )

    print("Fuse Eval: Done")
    return results


######################################
# Run
######################################
if __name__ == "__main__":
    # allocate gpus
    NUM_GPUS = 1
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    GPU.choose_and_enable_multiple_gpus(NUM_GPUS, force_gpus=force_gpus)

    RUNNING_MODES = ["train", "infer", "eval"]  # Options: 'train', 'infer', 'eval'

    # train
    if "train" in RUNNING_MODES:
        run_train(paths=PATHS, train_common_params=TRAIN_COMMON_PARAMS)

    # infer
    if "infer" in RUNNING_MODES:
        run_infer(paths=PATHS, infer_common_params=INFER_COMMON_PARAMS)

    # eval
    if "eval" in RUNNING_MODES:
        run_eval(paths=PATHS, eval_common_params=EVAL_COMMON_PARAMS)



"""
(C) Copyright 2021 IBM Corp.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
   http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Created on June 30, 2021
"""

import os
from typing import Hashable, Sequence
from fuse.data.datasets.dataset_base import DatasetBase
from fuse.utils.file_io.file_io import load_pickle, save_pickle
from sklearn.utils import shuffle
import pandas as pd
import numpy as np
from fuse.data.utils.export import ExportDataset
from fuse.data.utils.sample import get_sample_id_key


def print_folds_stat(db: pd.DataFrame, nfolds: int, key_columns: np.ndarray):
    """
    Print fold statistics
    :param db:                 dataframe which contains the fold partition
    :param nfolds:             Number of folds to divide the data
    :param key_columns:        keys for which balancing is forced
    """
    result = ""
    for f in range(nfolds):
        for key in key_columns:
            result += "----------fold" + str(f) + "\n"
            result += "key: " + key + "\n"
            result += db[db["fold"] == f][key].value_counts().to_string() + "\n"
    return result


def balanced_division(
    df: pd.DataFrame,
    no_mixture_id: str,
    keys_to_balance: Sequence[str],
    nfolds: int,
    seed: int = 1357,
    excluded_samples: Sequence[Hashable] = tuple(),
    print_flag: bool = False,
    debug_mode: bool = False,
) -> pd.DataFrame:
    """
    Partition the data into folds while using no_mixture_id for which no mixture between folds should be forced.
    and using keys_to_balance as the keys for which balancing is forced.
    The functions creates ID level labeling which is the cross-section of all possible mixture of key columns for that id
    it creates the folds so each fold will have about same proportions of ID level labeling while each ID will appear only in one fold
    For example - patient with ID 1234 has 2 images , each image has a binary classification (benign / malignant) .
    it can be that both of his images are benign or both are malignant or one is benign and the other is malignant.
    :param df:                 dataframe containing all samples including id and keys_to_balance
    :param no_mixture_id:      The key column for which no mixture between folds should be forced
    :param keys_to_balance:        keys for which balancing is forced
    :param nfolds:              number of folds to divide the data
    :param seed:               random seed used for creating folds
    :param excluded_samples:   sampled id which we do not want to include in the folds
    :param print_flag:         boolean flag which indicates if to print fold statistics
    """
    id_level_labels = []
    record_labels = []
    for field in keys_to_balance:
        values = df[field].unique()
        for value in values:
            value2 = str.replace(str(value), "+", "")
            # creates a binary label for each record and label
            record_key = "is" + value2
            df[record_key] = df[field] == value
            # creates a binary label for each id and label ( is anyone with this id has his label)
            id_level_key = "sample_id_" + field + "_" + value2
            df[id_level_key] = df.groupby([no_mixture_id])[record_key].transform(sum) > 0
            id_level_labels.append(id_level_key)
            record_labels.append(record_key)

    # drop duplicate id records
    samples_col = [no_mixture_id] + [col for col in id_level_labels]
    df_samples = df[samples_col].drop_duplicates()

    # generates a new label for each id based on sample_id value, using id's which are not in excluded_samples
    excluded_samples_df = df_samples[no_mixture_id].isin(excluded_samples)
    included_samples_df = df_samples[id_level_labels][~excluded_samples_df]
    df_samples["y_class"] = [str(t) for t in included_samples_df.values]
    y_values = list(df_samples["y_class"].unique())

    # initialize folds to empty list of ids
    db_samples = {}
    for f in range(nfolds):
        db_samples["data_fold" + str(f)] = []

    # creates a dictionary with key=fold , and values = ID which is in the fold
    # the partition goes as following : for each id level labels we shuffle the ID's and split equally ( as possible) to nfolds
    for y_value in y_values:
        patients_w_value = list(df_samples[no_mixture_id][df_samples["y_class"] == y_value])
        patients_w_value_shuffled = shuffle(patients_w_value, random_state=seed)
        splitted_array = np.array_split(patients_w_value_shuffled, nfolds)
        for f in range(nfolds):
            db_samples["data_fold" + str(f)] = db_samples["data_fold" + str(f)] + list(splitted_array[f])

    # creates a dictionary of dataframes, each dataframes holds all records for the fold
    # each ID appears only in one fold
    db = {}
    for f in range(nfolds):
        fold_df = df[df[no_mixture_id].isin(db_samples["data_fold" + str(f)])].copy()
        fold_df["fold"] = f
        db["data_fold" + str(f)] = fold_df
    folds = pd.concat(db, ignore_index=True)
    if print_flag is True:
        print_folds_stat(folds, nfolds, keys_to_balance)
    # remove labels used for creating the partition to folds
    if not debug_mode:
        folds.drop(id_level_labels + record_labels, axis=1, inplace=True)
    return folds


def dataset_balanced_division_to_folds(
    dataset: DatasetBase,
    output_split_filename: str,
    keys_to_balance: Sequence[str],
    nfolds: int,
    id: str = get_sample_id_key(),
    reset_split: bool = False,
    workers: int = 10,
    mp_context: str = None,
    **kwargs
):

    """
    Split dataset to folds.
    Support balancing, exclusion and radom seed (with a small improvement could support no mixture criterion).
    :param dataset: FuseMedML style dataset implementation.
    :param output_split_filename: filename to save/read the split from. If the file exist and reset_split=False - this function will return the split stored in reset_split.
    :param keys_to_balance: balancing any possible combination of values. For example for ["data.gender", "data.cancer"], the algorithm will balance each one of the following groups between the folds.
                            (gender=male, cancer=True), (gender=male, cancer=False), (gender=female, cancer=True), (gender=female, cancer=False)

    :param  nfolds : number of folds
    :param  id  : id to balance the split by ( not allowed 2 in same fold)
    :param reset_split: delete output_split_filename and recompute the split
    :param workers : numbers of workers for multiprocessing (eport dataset into dataframe)
    :param mp_context : multiprocessing context: "fork", "spawn", etc.
    :param kwargs: more arguments controlling the split. See function balanced_division() for details
    """
    if os.path.exists(output_split_filename) and not reset_split:
        return load_pickle(output_split_filename)
    else:
        if id == get_sample_id_key():
            keys = [get_sample_id_key()]
        else:
            keys = [get_sample_id_key(), id]
        if keys_to_balance is not None:
            keys += keys_to_balance
        df = ExportDataset.export_to_dataframe(dataset, keys, workers=workers, mp_context=mp_context)
        df_folds = balanced_division(df, id, keys_to_balance, nfolds, **kwargs)

        folds = {}
        for fold in range(nfolds):
            folds[fold] = list(df_folds[df_folds["fold"] == fold][get_sample_id_key()])
        save_pickle(folds, output_split_filename)
        return folds



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import os
import unittest

import random

import torchvision
from torchvision import transforms
from fuse.utils.rand.seed import Seed
from fuse.utils.ndict import NDict

import tempfile
from fuse.data.datasets.dataset_wrap_seq_to_dict import DatasetWrapSeqToDict


class TestDatasetWrapSeqToDict(unittest.TestCase):
    """
    Test sample caching
    """

    def setUp(self):
        pass

    def test_dataset_wrap_seq_to_dict(self):
        Seed.set_seed(1234)
        path = tempfile.mkdtemp()

        # Create dataset
        transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))])

        mnist_data_path = os.environ.get("MNIST_DATA_PATH", tempfile.mkdtemp(prefix="mnist"))
        torch_train_dataset = torchvision.datasets.MNIST(
            mnist_data_path, download=True, train=True, transform=transform
        )
        # wrapping torch dataset
        train_dataset = DatasetWrapSeqToDict(
            name="train", dataset=torch_train_dataset, sample_keys=("data.image", "data.label")
        )
        train_dataset.create()

        # get value
        index = random.randint(0, len(train_dataset))
        sample = train_dataset[index]
        item = torch_train_dataset[index]

        self.assertTrue(isinstance(sample, dict))
        self.assertTrue("data.image" in sample)
        self.assertTrue("data.label" in sample)
        self.assertTrue((sample["data.image"] == item[0]).all())
        self.assertEqual(sample["data.label"], item[1])

    def test_dataset_cache(self):
        Seed.set_seed(1234)

        transform = transforms.Compose([transforms.Normalize((0.1307,), (0.3081,))])
        # Create dataset
        mnist_data_path = os.environ.get("MNIST_DATA_PATH", tempfile.mkdtemp(prefix="mnist"))
        torch_dataset = torchvision.datasets.MNIST(mnist_data_path, download=True, train=True, transform=None)
        print(f"torch dataset size = {len(torch_dataset)}")

        # wrapping torch dataset
        cache_dir = tempfile.mkdtemp()

        dataset = DatasetWrapSeqToDict(
            name="test", dataset=torch_dataset, sample_keys=("data.image", "data.label"), cache_dir=cache_dir
        )
        dataset.create()

    def tearDown(self):
        pass


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
import logging
from typing import Callable, Dict, List, Sequence, Tuple

from fuse.utils import NDict

import torch
import numpy as np


class CollateToBatchList(Callable):
    """
    Collate list of dictionaries to a batch dictionary, each value in dict will be a list of values collected from all samples.
    """

    def __init__(
        self,
        skip_keys: Sequence[str] = tuple(),
        raise_error_key_missing: bool = True,
        special_handlers_keys: Dict[str, Callable] = None,
    ):
        """
        :param skip_keys: do not collect the listed keys
        :param raise_error_key_missing: if False, will not raise an error if there are keys that do not exist in some of the samples. Instead will set those values to None.
        """
        self._skip_keys = skip_keys
        self._raise_error_key_missing = raise_error_key_missing

    def __call__(self, samples: List[Dict]) -> Dict:
        """
        collate list of samples into batch_dict
        :param samples: list of samples
        :return: batch_dict
        """
        batch_dict = NDict()

        # collect all keys
        keys = self._collect_all_keys(samples)

        # collect values
        for key in keys:

            # skip keys
            if key in self._skip_keys:
                continue

            try:
                # collect values into a list
                collected_values, has_error = self._collect_values_to_list(samples, key)
                batch_dict[key] = collected_values
            except:
                print(f"Error: Failed to collect key {key}")
                raise

        return batch_dict

    def _collect_all_keys(self, samples: List[Dict]):
        """
        collect list of keys used in any one of the samples
        :param samples: list of samples
        :return: list of keys
        """
        keys = set()
        for sample in samples:
            if not isinstance(sample, NDict):
                sample = NDict(sample)
            keys |= set(sample.keypaths())
        return keys

    def _collect_values_to_list(self, samples: List[str], key: str) -> Tuple[List, bool]:
        """
        collect values of given key into a list
        :param samples: list of samples
        :param key: key to collect
        :return: list of values
        """
        has_error = False
        collected_values = []
        for index, sample in enumerate(samples):
            sample = NDict(sample)
            if key not in sample:
                has_error = True
                if self._raise_error_key_missing:
                    raise Exception(f"Error: key {key} does not exist in sample {index}: {sample}")
                else:
                    value = None
            else:
                value = sample[key]
            collected_values.append(value)
        return collected_values, has_error


def uncollate(batch: Dict) -> List[Dict]:
    """
    Reverse collate method
    Gets a batch_dict and convert it back to list of samples
    """
    samples = []
    if not isinstance(batch, NDict):
        batch = NDict(batch)
    keys = batch.keypaths()

    # empty batch
    if not keys:
        return samples

    batch_size = None
    for key in keys:
        if isinstance(batch[key], torch.Tensor):
            batch_size = len(batch[key])
            break

    if batch_size is None:
        for key in keys:
            if isinstance(batch[key], (np.ndarray, list, tuple)):
                batch_size = len(batch[key])
                break

    if batch_size is None:
        return batch  # assuming batch dict with no samples

    for sample_index in range(batch_size):
        sample = NDict()
        for key in keys:
            if isinstance(batch[key], (np.ndarray, torch.Tensor, list, tuple)):
                try:
                    sample[key] = batch[key][sample_index]
                except IndexError:
                    logging.error(f"Error - IndexError - key={key}, batch_size={batch_size}, len={batch[key]}")
                    raise
            else:
                sample[key] = batch[key]  # broadcast single value for all batch

        samples.append(sample)

    return samples



from fuse.data.ops.ops_common import OpApplyTypes
from fuseimg.utils.typing.key_types_imaging import type_detector_imaging
from functools import partial

OpApplyTypesImaging = partial(
    OpApplyTypes,
    type_detector=type_detector_imaging,
)



import unittest

from fuse.utils.ndict import NDict
from typing import Any, Union, List
import copy
from unittest.case import expectedFailure

from fuse.data.ops.op_base import OpBase, OpReversibleBase
from fuse.data.pipelines.pipeline_default import PipelineDefault


class OpSetForTest(OpReversibleBase):
    def __call__(self, sample_dict: NDict, op_id: str, key: str, val: Any) -> Union[None, dict, List[dict]]:
        # store information for reverse operation
        sample_dict[f"{op_id}.key"] = key
        if key in sample_dict:
            prev_val = sample_dict[key]
            sample_dict[f"{op_id}.prev_val"] = prev_val

        # set
        sample_dict[key] = val
        return sample_dict

    def reverse(self, sample_dict: NDict, op_id: str, key_to_reverse: str, key_to_follow: str) -> dict:
        key = sample_dict[f"{op_id}.key"]
        if key == key_to_follow:
            if f"{op_id}.prev_val" in sample_dict:
                prev_val = sample_dict[f"{op_id}.prev_val"]
                sample_dict[key_to_reverse] = prev_val
            else:
                if key_to_reverse in sample_dict:
                    sample_dict.pop(key_to_reverse)
        return sample_dict


class OpNoneForTest(OpBase):
    def __call__(self, sample_dict: NDict, **kwargs) -> Union[None, dict, List[dict]]:
        return None


class OpSplitForTest(OpBase):
    def __call__(self, sample_dict: NDict, **kwargs) -> Union[None, dict, List[dict]]:
        sample_id = sample_dict["data.sample_id"]
        samples = []
        split_num = 10
        for index in range(split_num):
            sample = copy.deepcopy(sample_dict)
            sample["data.sample_id"] = (sample_id, index)
            samples.append(sample)

        return samples


class TestPipelineDefault(unittest.TestCase):
    def test_pipeline(self):
        """
        Test standard backward and forward pipeline
        """
        pipeline_seq = [
            (OpSetForTest(), dict(key="data.test_pipeline", val=5)),
            (OpSetForTest(), dict(key="data.test_pipeline", val=6)),
            (OpSetForTest(), dict(key="data.test_pipeline_2", val=7)),
        ]
        pipe = PipelineDefault("test", pipeline_seq)

        sample_dict = NDict({})
        sample_dict = pipe(sample_dict)
        self.assertEqual(sample_dict["data.test_pipeline"], 6)
        self.assertEqual(sample_dict["data.test_pipeline_2"], 7)

        sample_dict = pipe.reverse(sample_dict, "data.test_pipeline", "data.test_pipeline")
        self.assertEqual("data.test_pipeline" in sample_dict, False)
        self.assertEqual(sample_dict["data.test_pipeline_2"], 7)

        sample_dict = pipe.reverse(sample_dict, "data.test_pipeline_2", "data.test_pipeline_2")
        self.assertEqual("data.test_pipeline" in sample_dict, False)
        self.assertEqual("data.test_pipeline_2" in sample_dict, False)

    def test_none(self):
        """
        Test pipeline with an op returning None
        """
        pipeline_seq = [
            (OpSetForTest(), dict(key="data.test_pipeline", val=5)),
            (OpNoneForTest(), dict()),
            (OpSetForTest(), dict(key="data.test_pipeline_2", val=7)),
        ]
        pipe = PipelineDefault("test", pipeline_seq)
        sample_dict = NDict({})
        sample_dict = pipe(sample_dict)
        self.assertIsNone(sample_dict)

    def test_split(self):
        """
        Test pipeline with an op splitting samples to multiple samples
        """
        pipeline_seq = [
            (OpSetForTest(), dict(key="data.test_pipeline", val=5)),
            (OpSplitForTest(), dict()),
            (OpSetForTest(), dict(key="data.test_pipeline_2", val=7)),
        ]
        pipe = PipelineDefault("test", pipeline_seq)
        sample_dict = NDict({"data": {"sample_id": 0}})
        sample_dict = pipe(sample_dict)
        self.assertTrue(isinstance(sample_dict, list))
        self.assertEqual(len(sample_dict), 10)
        expected_samples = [(0, i) for i in range(10)]
        samples = [sample["data.sample_id"] for sample in sample_dict]
        self.assertListEqual(expected_samples, samples)

    def tearDown(self) -> None:
        return super().tearDown()


if __name__ == "__main__":
    unittest.main()



from typing import Any, Dict, Sequence

import numpy as np

from pycocotools import mask as maskUtils


class MetricsInstanceSegmentaion:
    @staticmethod
    def convert_uncompressed_RLE_COCO_type(element: Dict, height: int, width: int) -> Dict:
        """
        converts uncompressed RLE to COCO default type ( compressed RLE)
        :param element:  input in uncompressed Run Length Encoding (RLE - https://en.wikipedia.org/wiki/Run-length_encoding)
                        saved in map object example :  {"size": [333, 500],"counts": [26454, 2, 651, 3, 13, 1]}
                        counts first element is how many bits are not the object, then how many bits are the object and so on
        :param height: original image height in pixels
        :param width: original image width in pixels
        :return  COCO default type ( compressed RLE)
        """
        p = maskUtils.frPyObjects(element, height, width)
        return p

    @staticmethod
    def convert_polygon_COCO_type(element: list, height: int, width: int) -> Dict:
        """
        converts polygon to COCO default type ( compressed RLE)
        :param element:  polygon - array of X,Y coordinates saves as X.Y , example: [[486.34, 239.01, 477.88, 244.78]]
        :param height: original image height in pixels
        :param width: original image width in pixels
        :return  COCO default type ( compressed RLE)
        """
        rles = maskUtils.frPyObjects(element, height, width)
        p = maskUtils.merge(rles)
        return p

    @staticmethod
    def convert_pixel_map_COCO_type(element: np.ndarray) -> Dict:
        """
        converts pixel map (np.ndarray) to COCO default type ( compressed RLE)
        :param element:  pixel map in np.ndarray with same shape as original image ( 0= not the object, 1= object)
        :return  COCO default type ( compressed RLE)
        """
        p = maskUtils.encode(np.asfortranarray(element).astype(np.uint8))
        return p

    @staticmethod
    def convert_to_COCO_type(input_list: Sequence[np.ndarray], height: int, width: int, segmentation_type: str) -> Dict:
        """
        converts all input list to COCO default type ( compressed RLE)
        :param input_list:  list of input in any COCO format
        :param height: original image height in pixels
        :param width: original image width in pixels
        :param segmentation_type: input format - pixel_map , uncompressed_RLE , compressed_RLE , polygon , bbox
        :return  list of output in COCO default type ( compressed RLE)
        """
        if segmentation_type == "uncompressed_RLE":
            output_list = [
                MetricsInstanceSegmentaion.convert_uncompressed_RLE_COCO_type(element, height, width)
                for element in input_list
            ]
        elif segmentation_type == "polygon":
            output_list = [
                MetricsInstanceSegmentaion.convert_polygon_COCO_type(element, height, width) for element in input_list
            ]
        elif segmentation_type == "pixel_map":
            output_list = [MetricsInstanceSegmentaion.convert_pixel_map_COCO_type(element) for element in input_list]

        return output_list

    @staticmethod
    def iou_jaccard(
        pred: Sequence[np.ndarray],
        target: Sequence[np.ndarray],
        segmentation_pred_type: str,
        segmentation_target_type: str,
        height: int,
        width: int,
    ) -> np.ndarray:
        """
        Compute iou (jaccard) score using pycocotools functions
        :param pred: sample prediction inputs - list of segmentations supported by COCO
        :param target: sample target inputs - list of segmentations supported by COCO
        :param segmentation_pred_type: input pred format - pixel_map , uncompressed_RLE , compressed_RLE , polygon , bbox
        :param segmentation_target_type: input target format - pixel_map , uncompressed_RLE ,polygon ( undergoes convertion to uncompressed_RLE ,see description above)
                                                               compressed_RLE , bbox - array in X.Y coordinates for example  [445.56, 205.16, 54.44, 71.55] ,
        :param height: height of the original image ( y axis)
        :param width: width of the original image ( x axis)
        :return matrix of iou computed between each element from pred and target
        """
        # convert input format to supported in pycocotools
        if segmentation_pred_type in ["pixel_map", "uncompressed_RLE", "polygon"]:
            pred = MetricsInstanceSegmentaion.convert_to_COCO_type(pred, height, width, segmentation_pred_type)
        if segmentation_target_type in ["pixel_map", "uncompressed_RLE", "polygon"]:
            target = MetricsInstanceSegmentaion.convert_to_COCO_type(target, height, width, segmentation_target_type)
        if segmentation_target_type == "uncompressed_RLE":
            iscrowd = list(np.ones(len(target)))
        else:
            iscrowd = list(np.zeros(len(target)))
        scores = maskUtils.iou(pred, target, iscrowd)
        return scores



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from distutils.log import warn
import unittest


from fuse.eval.examples.examples import (
    example_0,
    example_1,
    example_2,
    example_3,
    example_4,
    example_5,
    example_6,
    example_7,
    example_8,
    example_9,
    example_10,
    example_11,
    example_12,
    example_13,
)

from fuse.eval.examples.examples_segmentation import (
    example_seg_0,
    example_seg_1,
    example_seg_2,
    example_seg_3,
    example_seg_4,
)


class TestEval(unittest.TestCase):
    def test_eval_example_0(self):
        results = example_0()
        self.assertEqual(results["metrics.auc"], 0.845)

    def test_eval_example_1(self):
        results = example_1()
        self.assertEqual(results["metrics.auc"], 0.845)

    def test_eval_example_2(self):
        results = example_2()
        self.assertEqual(results["metrics.auc"], 0.845)
        self.assertAlmostEqual(results["metrics.auc_per_fold.std"], 0.0019, places=3)
        self.assertAlmostEqual(results["metrics.auc_per_fold.mean"], 0.839, places=3)
        self.assertAlmostEqual(results["metrics.auc_per_fold.median"], 0.839, places=3)
        self.assertAlmostEqual(results["metrics.auc_per_fold.0"], 0.841, places=3)
        self.assertAlmostEqual(results["metrics.auc_per_fold.1"], 0.837, places=3)

    def test_eval_example_3(self):
        results = example_3()
        self.assertAlmostEqual(results["metrics.auc_per_group.female"], 1.000, places=2)
        self.assertAlmostEqual(results["metrics.auc_per_group.male"], 0.666, places=2)

    def test_eval_example_4(self):
        results = example_4()
        self.assertEqual(results["metrics.auc.org"], 0.845)
        self.assertAlmostEqual(results["metrics.auc.mean"], 0.8456, places=3)
        self.assertAlmostEqual(results["metrics.auc.conf_lower"], 0.712, places=3)
        self.assertAlmostEqual(results["metrics.auc.conf_upper"], 0.954, places=3)
        self.assertAlmostEqual(results["metrics.auc.std"], 0.0622, places=3)

    def test_eval_example_5(self):
        results = example_5()
        self.assertAlmostEqual(
            results["metrics.compare_a_to_b.sensitivity.p_value"],
            1 - (0.33 + 0.67 * 0.33 + 0.67 * 0.67 * 0.33),
            places=2,
        )
        self.assertEqual(results["metrics.compare_a_to_b.sensitivity.count"], 10000)

    def test_eval_example_6(self):
        results = example_6()
        self.assertAlmostEqual(results["metrics.delongs_test.p_value.macro_avg"], 0.3173, places=4)
        self.assertAlmostEqual(results["metrics.delongs_test.z.macro_avg"], 1.000, places=3)

    def test_eval_example_7(self):
        results = example_7()
        self.assertAlmostEqual(results["metrics.delongs_test.p_value"], 0.09453, places=5)
        self.assertAlmostEqual(results["metrics.delongs_test.z"], 1.672, places=3)

    def test_eval_example_8(self):
        results = example_8()
        self.assertAlmostEqual(results["metrics.auc.macro_avg"], 0.808, places=3)
        self.assertAlmostEqual(results["metrics.auc.B"], 0.877, places=3)
        self.assertAlmostEqual(results["metrics.auc.VHR"], 0.723, places=3)

    def test_eval_example_9(self):
        try:
            import fuse.data
            import torchvision
        except ImportError:
            warn(" test_eval_example_8: requires fuse-med-ml-data and torchvision packages")
            return

        results = example_9()
        self.assertAlmostEqual(results["metrics.accuracy"], 0.1, places=2)

    def test_eval_example_10(self):
        results = example_10()
        self.assertAlmostEqual(results["metrics.mcnemars_test.statistic"], 1.0, places=1)
        self.assertAlmostEqual(results["metrics.mcnemars_test.p_value"], 1.0, places=5)

    def test_eval_example_11(self):
        results = example_11()
        self.assertEqual(results["metrics.accuracy"], 0.5)

    def test_eval_example_12(self):
        results = example_12()
        self.assertAlmostEqual(results["metrics.acc"], 0.615, places=2)

    def test_eval_example_seg_0(self):
        results = example_seg_0()
        self.assertAlmostEqual(results["metrics.dice_label_1.1"], 0.8740, places=3)
        self.assertAlmostEqual(results["metrics.dice_label_2.1"], 0.7163, places=3)
        self.assertAlmostEqual(results["metrics.dice_label_3.1"], 0.7184, places=3)

    def test_eval_example_seg_1(self):
        results = example_seg_1()
        self.assertAlmostEqual(results["metrics.dice.1"], 0.49900, places=3)
        self.assertAlmostEqual(results["metrics.overlap.1"], 0.51759, places=3)
        self.assertAlmostEqual(results["metrics.pixel_accuracy.1"], 0.48167, places=3)
        self.assertAlmostEqual(results["metrics.iou_jaccard.1"], 0.3324, places=3)

    def test_eval_example_seg_2(self):
        results = example_seg_2()
        self.assertAlmostEqual(results["metrics.dice.1"], 0.9726, places=3)
        self.assertAlmostEqual(results["metrics.dice.2"], 0.7184, places=3)
        self.assertAlmostEqual(results["metrics.dice.average"], 0.8455, places=3)

    def test_eval_example_seg_3(self):
        results = example_seg_3()
        self.assertEqual(results["metrics.hausdorff.average"], 1.0)
        self.assertAlmostEqual(results["metrics.dice.average"], 0.6667, places=3)
        self.assertAlmostEqual(results["metrics.overlap.average"], 1.0, places=3)
        self.assertAlmostEqual(results["metrics.pixel_accuracy.average"], 0.5, places=3)
        self.assertAlmostEqual(results["metrics.iou_jaccard.average"], 0.5, places=3)

    def test_eval_example_seg_4(self):
        results = example_seg_4()
        self.assertAlmostEqual(results["metrics.iou_bbox_person"], 0.81210, places=3)
        self.assertAlmostEqual(results["metrics.iou_bbox_car"], 0.895, places=3)
        self.assertAlmostEqual(results["metrics.recall_bbox_person"], 0.6875, places=3)
        self.assertAlmostEqual(results["metrics.recall_bbox_car"], 0.9318, places=3)
        self.assertAlmostEqual(results["metrics.precision_bbox_person"], 1.0, places=3)
        self.assertAlmostEqual(results["metrics.iou_polygon_car"], 0.8776, places=3)
        self.assertAlmostEqual(results["metrics.recall_polygon_person"], 0.60416, places=3)
        self.assertAlmostEqual(results["metrics.recall_polygon_car"], 0.8068, places=3)
        self.assertAlmostEqual(results["metrics.precision_polygon_person"], 0.875, places=3)
        self.assertAlmostEqual(results["metrics.precision_polygon_car"], 0.875, places=3)

    def test_eval_example_13(self):
        results = example_13()
        self.assertAlmostEqual(results["metrics.reliability"]["avg_accuracy"], 0.566, places=2)
        self.assertAlmostEqual(results["metrics.reliability"]["avg_confidence"], 0.441, places=2)
        self.assertAlmostEqual(results["metrics.ece"]["ece"], 0.126, places=2)
        self.assertAlmostEqual(results["metrics.ece"]["mce"], 0.800, places=2)
        self.assertAlmostEqual(results["metrics.find_temperature"], 0.8076, places=3)
        self.assertAlmostEqual(results["metrics.ece_calibrated"]["ece"], 0.123, places=2)
        self.assertAlmostEqual(results["metrics.ece_calibrated"]["mce"], 0.300, places=2)
        self.assertAlmostEqual(results["metrics.reliability_calibrated"]["avg_accuracy"], 0.566, places=2)
        self.assertAlmostEqual(results["metrics.reliability_calibrated"]["avg_confidence"], 0.485, places=2)


if __name__ == "__main__":
    unittest.main()
